#!/usr/bin/env python

"""pearsapp 1.7 - an RSS feed aggregator in (wx)Python

Copyright (C) 2003, 2004 Project 5

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

Contact info
web: http://come.to/project5
mail/msn: project5@wanadoo.nl
icq: 84243714
jabber: project5@rhymbox.com
"""
__version__ = "1.7"


import pearsgui
from wxPython import wx
from wxPython.lib.dialogs import wxScrolledMessageDialog
import sys, time
from wxPython.lib.mixins.listctrl import wxColumnSorterMixin
import future, htmldiff
import EmbeddedHTTPServer, BaseHTTPServer, httplib, ServerMixin
import copy
from pearsplugin import BasePlugin

try:
    import pearsdebug
    if pearsdebug.HAMSTERMODE:
        import hamsterengine as pearsengine
    else:
        import pearsengine
except:
    import hamsterengine as pearsengine # hamster engine is default
# Some options (note that these are stored and loaded at startup from
# the settings file. These are just initial values!)
MILLISECONDSREAD = 250 # number of milliseconds until item marked as read

from pearsconstants import *

ID_READTIMER = pearsgui.ni() # timer deciding when to toggle read/unread
ID_UPDATETIMER = pearsgui.ni() # timer deciding when to autoupdate

BROWSERDEFAULT = "<<OPEN_WITH_DEFAULT %URL>>"

# some global stuff; will be filled in by the main() function
Pears = None # main app
engine = None # the engine
mw = None # main window
url = None # url opener

def message(caption, text, style):
    """Wraps wx.wxMessageDialog: shows a message using the caption, text and style."""
    dlg = wx.wxMessageDialog(None, text, caption, style)
    result = dlg.ShowModal()
    dlg.Destroy()
    return result

def alert(caption, text):
    """Wraps message(): shows an alert message using the caption and text."""
    return message(caption, text, wx.wxOK|wx.wxICON_EXCLAMATION)
    
def ask(caption, text):
    """Wraps message(): asks the question text in a dialog with caption caption.
       Returns YES or NO (in pearsconstants.py) based on the answer."""
    ans = message(caption, text, wx.wxYES_NO|wx.wxICON_QUESTION)
    if ans == wx.wxID_YES:
        return YES
    else:
        return NO

def dispify(text):
    """Converts text from unicode to DISPCODE."""
    if type(text) == type("s"):
        if DISPCODE:
            # string to string
            return text
        else:
            # string to unicode.
            return text.decode(FEEDCODE, 'replace')
    elif type(text) == type(u"s"):
        if DISPCODE:
            # unicode to string
            return text.encode(DISPCODE, 'replace')
        else:
            # unicode to unicode
            return text
    else:
        if DEBUGMODE: print 'BIG ERROR:', `type(text)`
            
def SetHTMLFonts(htmlwindow, defaultfont):
    """Sets the HTML fonts of a wxHtmlWindow"""
    savedfont = engine.getSetting(FONT, {FFACE: defaultfont.GetFaceName(),
                                     FSIZE: defaultfont.GetPointSize(),
                                     FFAMILY: defaultfont.GetFamily()})
    fs = savedfont[FSIZE]
    htmlwindow.SetFonts(savedfont[FFACE], '', 
                        [int(round(fs*0.8)), int(round(fs*0.9)), fs, int(round(fs*1.2)),
                         int(round(fs*1.35)), int(round(fs*1.50)), int(round(fs*1.70))])

class AddDialog(pearsgui.MyAddDialog):
    """Wraps class created by wxGlade in pearsgui. Returns
       0 if nothing has been added and 1 if a feed has been
       added."""
    def __init__(self, *args, **kwds):
        pearsgui.MyAddDialog.__init__(self, *args, **kwds)
            
        self.Centre()
            
        self.SetStatus = self.StatusLabel.SetLabel
            
        self.OkButton.SetLabel("Get")
        self.OkButton.SetToolTipString("Click to get feed info")
        self.OkButton.Enable(False)
        self.AddButton.Enable(False)
        self.SetStatus(" ")
        wx.EVT_BUTTON(self, pearsgui.ID_ADDOK, self.doGet)
        wx.EVT_BUTTON(self, pearsgui.ID_ADDCANCEL, self.doClose)
        wx.EVT_BUTTON(self, pearsgui.ID_ADDADD, self.doAdd)
        wx.EVT_TEXT(self, pearsgui.ID_ADDLOCATION, self.doEnable)
        
    def doEnable(self, event):
        """Enables or disables the OK button based on whether
           the URL is already in the list of URLs."""
        alreadyexists = self.FeedLocation.GetValue().strip() in engine.getFeedsLocations()
        nofeed = self.FeedLocation.GetValue().strip()==""
        if nofeed or alreadyexists:
            self.OkButton.Enable(False)
        else:
            self.OkButton.Enable(True)
        if alreadyexists:
            self.SetStatus("This feed already exists in your feeds list.")
        elif nofeed:
            self.SetStatus("Please specify the location of the feed you wish to add.")
        else:
            self.SetStatus("Click on the Get button to check if feed is correct.")
        
    def doClose(self, event):
        """Close the modal dialog"""
        self.EndModal(0)
        
    def doAdd(self, event):
        """After a valid RSS feed has been received with doGet(), doGet()
           will set the BUTTON event of the OK button to point to to doAdd(), 
           so clicking OK will add this feed to the list."""
        if DEBUGMODE: 
            print "[Debug]: trying to add feed in the engine"
            print "         location=%s" % self.FeedLocation.GetValue().strip()
            print "         title=%s" % `self.FeedTitle.GetValue().strip()`
            if DISPCODE:
                print "         title=%s" % self.FeedTitle.GetValue().strip().encode(DISPCODE, "replace")
        # determine autorefresh interval; this is a text which must be converted to a number
        # this interval is given in minutes
        up = self.AutoRefresh.GetStringSelection()
        if up.lower().strip() == "never":
            up = NEVER # very large number :) equal to about 20 years
        else:
            up = up.lower().strip().split()
            # up is now a list of the form [<number>, <minutes|hours>, <optional:other stuff>]
            if up[1].count("hour")>0:
                multiplier = 60 # convert to minutes
            else:
                multiplier = 1
            try:
                up = int(up[0])*multiplier
            except:
                up = NEVER
            
        ok = engine.addFeed(location=self.FeedLocation.GetValue().strip(),
                            title=self.FeedTitle.GetValue().strip(),
                            cachesize=int(self.TopicsStore.GetStringSelection()),
                            updateinterval=up,
                            channel=self.feedprops)
        if DEBUGMODE:
            print "[Debug]: adding succeeded? (1=True, 0=False): %s" % ok
        self.EndModal(1)
        
    def prepareForAdd(self):
        """Modify interface to switch to Add mode"""
        self.FeedLocation.SetEditable(False)
        self.OkButton.Enable(False)
        self.AddButton.Enable(True)
        self.AddButton.SetDefault()            
        
    def showProps(self):
        """Display the feed properties in the appropriate field."""
        t = []
        if self.feedprops.has_key('tagline'): # if feed has description
            t.append(self.feedprops['tagline'])
            t.append(u"")
            
        sortedkeys = self.feedprops.keys()
        sortedkeys.sort()
        for key in sortedkeys:
            print key
            if self.feedprops[key] and \
              not key in [u"description", u"title", u"link", u"guid", u"tagline", u"items"]:
                if type(self.feedprops[key]) in [type(u""), type("")]:
                    t.append(u": ".join([key, engine.htmlToPlain(self.feedprops[key], DISPCODE)]))
        self.FeedDescription.SetValue(dispify(u"\n".join(t)))
        
    def doGet(self, event):
        """Try to get feed properties from the feed location."""
        self.feedprops = {}
        try:
            ok = True
            self.FeedDescription.Clear()
            self.DescriptionLabel.SetLabel("Properties")
                
            self.feedprops = engine.downloadChannelProperties(self.FeedLocation.GetValue().strip())
            
            if len(str(self.feedprops))<5: # not downloaded
                ok = False
        except:
            import traceback
            if DEBUGMODE: traceback.print_exc()
            ok = False
        
        if ok:
            # add Title from properties if not user-specified
            if not self.FeedTitle.GetValue().strip():
                if self.feedprops.has_key('title'): # if feed has title
                    self.FeedTitle.SetValue(engine.htmlToPlain(self.feedprops['title'], DISPCODE))
                else: # if feed has no title specified
                    self.FeedTitle.SetValue(self.FeedLocation.GetValue())
            self.showProps()
            self.prepareForAdd()
            self.SetStatus("Click on the Add button to add this feed")
        else:
            if not self.feedprops:
                self.SetStatus("Error: this feed was either not found or is of an invalid format")
            else:
                self.FeedDescription.SetValue("Feed properties = "+str(self.feedprops))
    

class AboutDialog(pearsgui.MyAboutDialog):
    """Wraps class created by wxGlade in pearsgui."""
    def __init__(self, *args, **kwds):
        pearsgui.MyAboutDialog.__init__(self, *args, **kwds)
        
        self.AboutText.SetWindowStyle(wx.wxSUNKEN_BORDER)
        self.Centre()
        global url
        self.AboutText.urlOpener = url
        
        SetHTMLFonts(self.AboutText, self.GetFont())
        
        # load the HTML
        fname = os.path.join(PEARSLOCATION, "about.html.dat")
        try:
            textfile = file(fname, "r")
            text = textfile.read()
            textfile.close()
        except:
            text = "<b>Could not load data file</b>:<br>%s" % fname
        feedsnumber = len(engine.getFeedsNames())
        topicsnumber = engine.getTotalTopicsCount()
        feedssize = engine.getFeedsSize()
        topicssize = engine.getTotalTopicsSize()
        replacements = {"pearsversion": __version__, 
                        "programdir": PEARSLOCATION, 
                        "parserversion": engine.parserversion,
                        "datadir": engine.savedir, 
                        "pythonversion": PYTHONVERSION,
                        "wxpythonversion": WXPYTHONVERSION,
                        "timesexecuted": engine.getSetting(TIMESRUN, 0),
                        "feedsnumber": feedsnumber,
                        "topicsnumber": topicsnumber,
                        "feedssize": "%.1f" % (feedssize/1024.0),
                        "topicssize": "%.1f" % (topicssize/1024.0),
                        "averagefeedsize": "%.1f" % ((feedssize+topicssize)/feedsnumber/1024.0),
                        "sessionuptime": mw.uptimeToStr(mw.getSessionUptime()),
                        "totaluptime": mw.uptimeToStr(mw.getTotalUptime()),
                        "enginetype": pearsengine.__version__,}
        text = text % replacements
        self.AboutText.SetPage(text)
        

class ImpexDialog(pearsgui.MyImpexDialog, wxColumnSorterMixin):
    """Wraps class created by wxGalde in pearsgui."""
    def __init__(self, *args, **kwds):
        pearsgui.MyImpexDialog.__init__(self, *args, **kwds)
        self.feeds = []
        self.Centre()
        self.SetTitle(kwds["title"])
        
        self.itemDataMap = {} # for the wxColumnSorterMixin; also used for returning selected items
        wxColumnSorterMixin.__init__(self, 3)
        
        # initiate controls
        self.ImpexFilter.SetValue("")
        # add columns to list
        self.ImpexList.InsertColumn(0, "Title")
        self.ImpexList.InsertColumn(1, "Description")
        self.ImpexList.InsertColumn(2, "Location")
     
        # make sure list is re-evaluated whenever the filter changes
        wx.EVT_TEXT(self, pearsgui.ID_IMPEXFILTER, self.drawItems)
        wx.EVT_LIST_ITEM_SELECTED(self, pearsgui.ID_IMPEXLIST, self.updateStatus)
        wx.EVT_LIST_ITEM_DESELECTED(self, pearsgui.ID_IMPEXLIST, self.updateStatus)
        wx.EVT_BUTTON(self, wx.wxID_CANCEL, self.doClose)
        wx.EVT_BUTTON(self, pearsgui.ID_IMPEXSELECTALL, self.selectAll)
        wx.EVT_BUTTON(self, pearsgui.ID_IMPEXDESELECTALL, self.deselectAll)
        
    def setOkButtonLabel(self, text="OK"):
        """Modifies the label of the Ok button to text"""
        self.OkButton.SetLabel(text)
        
    def setItems(self, feedslist):
        """Lists the feeds in the feedslist (these items
           must be: on index 0 the address, on index 1 the
           title (or name) and on index 2 the description)."""
        self.feeds = feedslist
        for index in range(len(self.feeds)):
            # for the wxColumnSorterMixin
            self.itemDataMap[index] = feedslist[index]
        self.drawItems()
        
    def GetListCtrl(self):
        """Necessary for the ColumnSorter Mixin"""
        return self.ImpexList
        
    def drawItems(self, event=None):
        """Draws the items, using the filter."""
        self.ImpexList.DeleteAllItems()
        
        # Rebuild self.itemDataMap for the wxColumnSorterMixin.
        # This is done by applying the filter to self.feeds.
        # 1. Prepare for filtering
        self.itemDataMap = {}
        filter = self.ImpexFilter.GetValue().strip().upper()
        index = 0
        for data in self.feeds:
            item = [ field.strip().upper() for field in data ]
            # 2. Filter to determine if item should be drawn:
            #    this list will be empty if no match for the filter is found
            #    being empty, it will also be evaluated as False
            fitsfilter = []
            for field in item:
                _field, _filter = field, filter
                if type(field)!=type(filter):
                    if type(field)==type(""):
                        _filter = filter.encode("ascii", "replace")
                    else:
                        _field = field.encode("ascii", "replace")
                if _field.find(_filter)>=0:
                    fits = True
                else:
                    fits = False
                if fits:
                    fitsfilter.append(field) # no further encoding necessary
            if filter=="" or fitsfilter:
                self.itemDataMap[index] = data
                index+=1
            
        # the data in self.itemDataMap is now filtered an can be shown
        items = self.itemDataMap.items()
        for x in range(len(items)):
            key, data = items[x]
            self.ImpexList.InsertStringItem(x, dispify(data[1]))
            self.ImpexList.SetStringItem(x, 1, dispify(data[2]))
            self.ImpexList.SetStringItem(x, 2, dispify(data[0]))
            self.ImpexList.SetItemData(x, key)
               
        self.updateStatus()
        self.SortListItems(col = 0)
        
        if not engine.getSetting(STORECOLUMNS, True):
            self.ImpexList.SetColumnWidth(0, wx.wxLIST_AUTOSIZE)
            self.ImpexList.SetColumnWidth(1, wx.wxLIST_AUTOSIZE)
            self.ImpexList.SetColumnWidth(2, wx.wxLIST_AUTOSIZE)
                
    def updateStatus(self, event=None):
        """Updates the status field"""
        template = "Current feeds status: %s total, %s visible, %s selected"
        self.StatusText.SetLabel(template % \
                                 (len(self.feeds), 
                                  self.ImpexList.GetItemCount(), 
                                  self.ImpexList.GetSelectedItemCount()))
                                  
    def deselectAll(self, event=None):
        """Selects all items in the list control."""
        for index in range(self.ImpexList.GetItemCount()):
            self.ImpexList.SetItemState(index, 0, wx.wxLIST_STATE_SELECTED)
        
    def selectAll(self, event=None):
        """Selects all items in the list control."""
        for index in range(self.ImpexList.GetItemCount()):
            self.ImpexList.SetItemState(index, wx.wxLIST_STATE_SELECTED, wx.wxLIST_STATE_SELECTED)
            
    def getSelection(self):
        """Returns the selected feeds as a list of location, 
           title and description tuples."""
        res = []
        il = self.ImpexList
        current = -1 # keeps track of where to start searching for more selected items
        while True:
            current = il.GetNextSelected(current)
            if current<0: 
                break
            data = self.itemDataMap[il.GetItemData(current)]
            res.append((data[0], data[1], data[2]))
        return res
            
    def doClose(self, event):
        """Close the modal dialog"""
        self.EndModal(0)

class urlOpener(object):
    """Opens urls."""
    def __init__(self):
        pass
    def open(self, url="http://come.to/project5", command=None):
        """Opens the url with the method defined in the settings.
           If command==None, it is retreived from the settings.
        """       
        if command==None:
            command = engine.getSetting(BROWSERCOMMAND, BROWSERDEFAULT)
        if command==BROWSERDEFAULT.strip():
            # use default opening method
            import os
            if os.environ.has_key("BROWSER") and \
               os.environ["BROWSER"]=='kfmclient openProfile webbrowsing':
                print "Invalid browser detected : %s\nResetting to konqueror." % os.environ["BROWSER"]
                os.environ["BROWSER"] = 'konqueror' # set it to konqueror
            import webbrowser # MUST be imported only AFTER os.environ has been modified
            webbrowser.open(url, 1)
        else: #otherwise use user-defined command line
            # replase %URL with actual url
            command = command.replace("%URL", url)
            import os
            os.system(command)

class BrowserDialog(pearsgui.MyBrowserDialog):
    """Wraps the class created by wxGlade in pearsgui."""
    def __init__(self, *args, **kwds):
        pearsgui.MyBrowserDialog.__init__(self, *args, **kwds)
        # set predefined command lines
        anyoscommands = {"Any OS: default browser (this is the default action)": BROWSERDEFAULT}
        platformcommands = {}
        platform = sys.platform
        if platform.find("win32")>=0:
            platformcommands = {"Windows: Internet Explorer": "start iexplore %URL",
                                "Windows: Mozilla": "start mozilla.exe %URL",
                                "Windows: Opera": "start opera.exe %URL"
                               }
        else:
            platformcommands = {"Linux: Galeon": "galeon %URL",
                                "Linux: Galeon, existing window": "galeon --existing %URL",
                                "Linux: Galeon, new tab": "galeon --new-tab %URL",
                                "Linux: Galeon, new tab, don't raise window": "galeon --new-tab --noraise %URL",
                                "Linux: Galeon, new window": "galeon --new-window --noraise %URL",
                                "Linux: Galeon, new window, don't raise window": "galeon --new-window %URL",
                                "Linux: Konqueror, new window": "konqueror --new-window %URL",
                                "Linux: Mozilla": """mozilla -remote "openUrl(%URL)" 2>/dev/null >/dev/null""",
                                "Linux: Mozilla, new tab": """mozilla -remote "openUrl(%URL, new-tab)" 2>/dev/null >/dev/null""",
                               }
        # concatenate commands
        self.commands = {}
        for keys in (anyoscommands.keys(), platformcommands.keys()):
            keys.sort()
            for key in keys: 
                self.commands[key] = (anyoscommands.has_key(key) and [anyoscommands[key]] or [platformcommands[key]])[0]
        # write combobox items
        bcc = self.BrowserCommandComboBox
        bcc.Clear()
        keys = self.commands.keys()
        keys.sort()
        for key in keys:
            bcc.Append(key)
        # set current command
        bcc.SetValue(engine.getSetting(BROWSERCOMMAND, BROWSERDEFAULT))
            
        self.bcc = bcc # less typing in the rest of the code
            
        self.verifyValidity()
            
        # bind events
        wx.EVT_BUTTON(self, pearsgui.ID_BROWSERTEST, self.doTest)
        wx.EVT_BUTTON(self, pearsgui.ID_BROWSEROK, self.apply)
        wx.EVT_COMBOBOX(self, pearsgui.ID_BROWSERCOMMANDBOX, self.loadCommand)
        wx.EVT_TEXT(self, pearsgui.ID_BROWSERCOMMANDBOX, self.verifyValidity) 
        
    def doTest(self, event):
        """Tests the new command line."""
        command = self.bcc.GetValue()
        url.open("http://come.to/project5", command)
        
    def apply(self, event):
        """Save new command line."""
        engine.setSetting(BROWSERCOMMAND, self.bcc.GetValue())
        self.EndModal(0)
        
    def loadCommand(self, event):
        """Loads the selected command line when an item is selected
           in the listbox."""
        key = event.GetString() # get new selection
        command = self.commands[key] # get mapped command
        wx.wxCallAfter(self.bcc.SetValue, command)
        wx.wxCallAfter(self.verifyValidity)
        
    def verifyValidity(self, event=None):
        """Checks whether the command line contains "%URL" and
           enables/disables buttons accordingly."""
        if self.bcc.GetValue().find("%URL")<0:
            # invalid, disable buttons
            self.OkButton.Enable(False)
            self.TestButton.Enable(False)
        else:
            # valid, enable buttons
            self.OkButton.Enable(True)
            self.TestButton.Enable(True)


class MainFrame(pearsgui.MyMainFrame, ServerMixin.ServerMixin):
    """Wraps class created by wxGlade in pearsgui."""
    def __init__(self, *args, **kwds):
        pearsgui.MyMainFrame.__init__(self, *args, **kwds)
        
        serv = BaseHTTPServer.HTTPServer(
                    ('localhost',RADIOPORT),
                    EmbeddedHTTPServer.EmbeddedHTTPRequestHandler,
               )
        # XXX make dummy final connection to force server thread to exit!
        servfin = lambda x='localhost',y=RADIOPORT: httplib.HTTP().connect(x,y)
        ServerMixin.ServerMixin.__init__(self, server=serv, serverfinish=servfin)
        EmbeddedHTTPServer.register('/system/pages/subscriptions',
                                    ServerMixin.notifyGUI, 
                                    self, 
                                    'addFeedRequest')
        
        # configure SplitterWindows
        self.MainSplitter.SetMinimumPaneSize(75)
        self.RightSplitter.SetMinimumPaneSize(75)
        self.MainSplitter.SetSashPosition(190)
        self.RightSplitter.SetSashPosition(150)
        self.SetTitle("Pears " + __version__)
        self.Centre()
            
        global url
        self.TopicText.urlOpener = url
            
        # set the icon
        if DEBUGMODE: print PEARSLOCATION
        iconfile = os.path.join(PEARSLOCATION, "pearssmall.ico")
        theicon = wx.wxIcon(iconfile, wx.wxBITMAP_TYPE_ICO)
        self.SetIcon(theicon)
        
        # reset contents of everything
        self.FeedsList.ClearAll()
        self.FeedsList.CurrentItem = None
        self.LockedFeed = None
        self.FeedTopics.ClearAll()
        self.FeedTopics.CurrentItem = None
        self.TopicText.SetPage("")
        
        # set help text
        self.FeedsList.SetHelpText("Click to open, right-click to update")
        self.FeedTopics.SetHelpText("Click to open")
        
        # enable displaying of the URL in the status bar
        self.TopicText.SetRelatedFrame(self, "%s")
        self.TopicText.SetRelatedStatusBar(0)
        
        # add columns to lists
        self.FeedsList.InsertColumn(0, "Title")
        self.FeedsList.InsertColumn(1, "New", wx.wxLIST_FORMAT_RIGHT)
        self.FeedTopics.InsertColumn(0, "Topic")
        
        # load the plugins
        self.Plugins = self.loadPlugins()
        
        # make the readtimer
        self.ReadTimer = wx.wxTimer(self, ID_READTIMER)
        # make the autoupdate timer and start it
        self.UpdateTimer = wx.wxTimer(self, ID_UPDATETIMER)
        self.UpdateTimer.Start(int(AUTOUPDRESOLUTION*60*1000), # AUTOUPDRESOLUTION is in minutes
                               wx.wxTIMER_CONTINUOUS)
        
        # allow URL dropping on the feeds list
        self.SetDropTarget(pearsgui.MyURLDropTarget(self))
        
        # bind some events
        wx.EVT_SIZE(self, self.doSize)
        
        wx.EVT_MENU(self, pearsgui.ID_EXIT, self.doExit)
        wx.EVT_MENU(self, pearsgui.ID_ADD, self.addFeed)
        wx.EVT_MENU(self, pearsgui.ID_REMOVE, self.removeFeed)
        wx.EVT_MENU(self, pearsgui.ID_SHOWFEEDPROPERTIES, self.showFeedProperties)
        wx.EVT_CLOSE(self, self.doClose)
        
        wx.EVT_MENU(self, pearsgui.ID_HELP, self.showHelp)
        wx.EVT_MENU(self, pearsgui.ID_ABOUT, self.showAbout)
        
        wx.EVT_LIST_ITEM_SELECTED(self, pearsgui.ID_FEEDSLIST, self.showFeedItems)
        wx.EVT_LIST_ITEM_RIGHT_CLICK(self, pearsgui.ID_FEEDSLIST, self.rclickFeed)
        wx.EVT_LIST_COL_CLICK(self, pearsgui.ID_FEEDSLIST, self.sortFeedsColumn)
        wx.EVT_MENU(self, pearsgui.ID_UPDATE, self.updateFeed)
            
        wx.EVT_MENU(self, pearsgui.ID_UPDATEALL, self.updateAll)
        wx.EVT_MENU(self, pearsgui.ID_UPDATEALLFORCE, self.updateAll)
        wx.EVT_TIMER(self, ID_UPDATETIMER, self.updateAll) 
                     # the implementation finds out on its own whether it was 
                     # called by the timer or by the menu item
            
        wx.EVT_MENU(self, pearsgui.ID_MARKALLFEEDSREAD, self.markAllFeedsRead)
        wx.EVT_MENU(self, pearsgui.ID_EMPTYALLFEEDS, self.emptyAllFeeds)
            
        ### Sorting actions
        # moveFeed figures out whether it's called by MOVEUP or MOVEDOWN
        wx.EVT_MENU(self, pearsgui.ID_MOVEUP, self.moveFeed)   
        wx.EVT_MENU(self, pearsgui.ID_MOVEDOWN, self.moveFeed) 
        # sortFeeds figurs out on its own by which menu item it was called
        wx.EVT_MENU(self, pearsgui.ID_SORTFEEDSALFA, self.sortFeeds) 
        wx.EVT_MENU(self, pearsgui.ID_SORTFEEDSINVALFA, self.sortFeeds)
        wx.EVT_MENU(self, pearsgui.ID_SORTFEEDSREVERSE, self.sortFeeds)
        wx.EVT_MENU(self, pearsgui.ID_SORTFEEDSRANDOM, self.sortFeeds)
        wx.EVT_MENU(self, pearsgui.ID_SORTFEEDSUNREAD, self.sortFeeds)
            
        wx.EVT_MENU(self, pearsgui.ID_IMPORT, self.importFeeds)
        wx.EVT_MENU(self, pearsgui.ID_EXPORT, self.exportFeeds)
            
        wx.EVT_LIST_ITEM_SELECTED(self, pearsgui.ID_FEEDTOPICS, self.showTopic)
        wx.EVT_LIST_ITEM_RIGHT_CLICK(self, pearsgui.ID_FEEDTOPICS, self.rclickTopic)
            
        wx.EVT_TIMER(self, ID_READTIMER, self.markTopicRead)
            
        wx.EVT_MENU(self, pearsgui.ID_GOTOLINK, self.gotoLink)
        wx.EVT_MENU(self, pearsgui.ID_MARKITEMREAD, self.markTopicRead)
        wx.EVT_MENU(self, pearsgui.ID_MARKITEMUNREAD, self.markTopicUnread)
        wx.EVT_MENU(self, pearsgui.ID_DELETEITEM, self.deleteTopic)
        wx.EVT_MENU(self, pearsgui.ID_MARKALLITEMSREAD, self.markAllTopicsRead)
        wx.EVT_MENU(self, pearsgui.ID_MARKALLITEMSUNREAD, self.markAllTopicsUnread)
        wx.EVT_MENU(self, pearsgui.ID_DELETEALLITEMS, self.deleteAllTopics)
        wx.EVT_MENU(self, pearsgui.ID_COPYTDL, self.copyTitleDescriptionLink)
        wx.EVT_MENU(self, pearsgui.ID_COPYLINK, self.copyLink)
        wx.EVT_MENU(self, pearsgui.ID_COPYALL, self.copyAll)
            
        wx.EVT_MENU(self, pearsgui.ID_REMOVEIMAGES, self.setRemoveImages)
        wx.EVT_MENU(self, pearsgui.ID_PARANOIDSAVING, self.setParanoidSaving)
        wx.EVT_MENU(self, pearsgui.ID_SHOWDIFF, self.setShowDiff)
        wx.EVT_MENU(self, pearsgui.ID_SMARTDISPLAY, self.setSmartDisplay)
        wx.EVT_MENU(self, pearsgui.ID_STORECOLUMNS, self.setStoreColumns)
        wx.EVT_MENU(self, pearsgui.ID_REVERSEORDER, self.setReverseOrder)
        wx.EVT_MENU(self, pearsgui.ID_BROWSER, self.setBrowser)
        wx.EVT_MENU(self, pearsgui.ID_FONTSELECT, self.setFont)
        wx.EVT_MENU(self, pearsgui.ID_TIMEREAD, self.setTimeRead)
        wx.EVT_MENU(self, pearsgui.ID_DUMPETAG, self.dumpEtag)
        wx.EVT_MENU(self, pearsgui.ID_LAUNCHSHELL, self.launchShell)
            
        wx.EVT_MENU(self, pearsgui.ID_RIGHTCLICKPEARS, self.setRightclick)
        wx.EVT_MENU(self, pearsgui.ID_RIGHTCLICKSYSTEM, self.setRightclick)
            
        wx.EVT_MENU(self, pearsgui.ID_PLUGINSREFRESH, self.loadPlugins)
        wx.EVT_MENU(self, pearsgui.ID_PLUGINWARN, self.setPluginWarn)
        
    def dumpEtag(self, event):
        """Remove all etag and modified properties from the feeds."""
        engine.dumpEtag()
        
    def runPlugin(self, event):
        """Event handler for plugin menu items. Runs the plugin that
           was clicked on."""
        # find out which plugin was clicked and what it's supposed to do
        m_id = event.GetId()
        plugin = self.pluginmapper[m_id]
        sourcefilename = plugin[PSource] # must be and REMAIN the file name!
        klass = plugin[PClass]
        if engine.getSetting(PLUGINWARN, True):
            dlg = wx.wxMessageDialog(self, 
                        "\n".join(
                                  ["You are trying to run the plugin class '%s' in file" % klass, 
                                   "   %s" % sourcefilename,
                                   "Plugins can have a very profound effect on Pears and your system",
                                   "in general, including very bad ones. You should not run plugins",
                                   "that you do not trust completely.", "",
                                   "Do you want to continue?"]),
                        "Confirm plugin execution",
                        wx.wxYES_NO|wx.wxICON_QUESTION)
            if dlg.ShowModal()!=wx.wxID_YES:
                return None
            
        # load the code from the source
        try:
            sf = file(sourcefilename, "r")
            code = sf.read()
            sf.close()
        except:
            alert("Plugin warning", "Could not read code from the file:\n   " % source)
            return None
        code = code.replace('\r\n', '\n') # \r\n is not allowed in code that will be compiled
        # run the code
        try:
            code = compile(code, sourcefilename, 'exec')
            exec code in globals()
            # create object from the klass specified in the plugin configuration
            pluginobject = eval("%s()" % klass)
            pluginobject.run()
        except: # show traceback
            sys.excepthook(sys.exc_info()[0], sys.exc_info()[1], sys.exc_info()[2])            
        
    def loadPlugins(self, event=None):
        """Loads all plugins from the plugins directory and adds them to
           the Plugins menu. Is also the event handler for the Refresh plugins
           menu item.
           
           The plugin system works as follows: 
           There are two lists: self.plugcolls (stores plugin collections) and
           self.plugs (stures individual plugins).
           
           Each item in plugcolls is a dictionary with the following keys:
           - LongName: the full name of the plugin (str)
           - ShortName: short name, shown in the menu (str)
           - Hint: hint shown in the menu (str)
           - Version: version of the plugin collection (str)
           - Author: name of the plugin collection author (str)
           - Plugins: list of plugins (list)
           - Source: file path where this collection (.pears) is located (str)
           Each item in the Plugins list is a dictionary with the following keys:
           - ShortName: short name, shown in the menu (str)
           - Hint: shown in the menu (str)
           - Class: class name of the plugin (str)
           - Source: same source as above (str) - used for convenience
           
           Each item in the plugs list is a dictionary with the following keys:
           - LongName, ShortName, Hint, Version, Author, Source, Class: as above
           """
        self.plugs = [] # individual plugins
        self.plugcolls = [] # plugin collections
        self.pluginmapper = {} # maps a menu id to a plugin dictionary which 
                               # has at least a "Class" (PClass) and a "Source" (PSource) key
        pluginsdir = os.path.join(PEARSLOCATION, 'plugins')
        try:
            files = os.listdir(pluginsdir)
        except:
            return None
        files = [ os.path.join(pluginsdir, f) for f in files ]
        files = [ f for f in files if f.endswith(PLUGINEXTENSION) and os.path.isfile(f) ]
        # files no contains only real plugin files, with their full path
        # read plugin data
        for filename in files:
            self.interpretPluginFile(filename)
        # clean up Plugins menu
        pluginsmenu = self.MenuBar.FindMenu("Plugins")
        pluginsmenu = self.MenuBar.GetMenu(pluginsmenu)
        # pluginsmenu now contains the plugins menu
        for mni in pluginsmenu.GetMenuItems()[2:]: # first two items must remain there
            pluginsmenu.Remove(mni.GetId())
            mni.Destroy()
        
        def addmenuitem(plug, targetmenu):
            """Function which adds a menu item for the plug plugin to 
               the targetmenu menu."""
            m_id = pearsgui.ni()
            self.pluginmapper[m_id] = plug # will be used by the event handler
            targetmenu.Append(m_id, plug[PShortName], plug[PHint])
            wx.EVT_MENU(self, m_id, self.runPlugin) # bind event handler
            
        # now add plugin menu items for the available plugins
        if self.plugs:
            pluginsmenu.AppendSeparator()
        # loop through the individual plugins
        for plugin in self.plugs:
            addmenuitem(plugin, pluginsmenu)
        # on to the plugin collections
        if self.plugcolls:
            pluginsmenu.AppendSeparator()
        for coll in self.plugcolls:
            # each collection has a menu-item with a submenu for the individual plugins
            # inside that collection
            collmenu = wx.wxMenu() 
            for plugin in coll[PPlugins]:
                addmenuitem(plugin, collmenu)
            pluginsmenu.AppendMenu(pearsgui.ni(), coll[PShortName], collmenu, coll[PHint])
        
    def interpretPluginFile(self, filename):
        """Reads the plugin file at <filename> and inserts any plugins it
           contains in self.plugs or self.plugcolls.
           Arguments:
             filename - path to the supposed plugin file"""
        pf = file(filename, "r")
        try:
            while 1: # look for the plugin definition
                t = pf.readline()
                if not t:
                    break # end of file reached
                t = t.strip()
                if t==PStartPlugin: # found a plugin definition
                    # start with default, empty plugin data
                    plugin = {PLongName: PUnknown,
                              PShortName: PUnknown,
                              PHint: PUnknown,
                              PVersion: PUnknown,
                              PAuthor: PUnknown,
                              PSource: filename,
                              PClass: PUnknown,
                                  }
                    while 1: # read single plugin
                        t = pf.readline().strip()
                        if not t or t==PEnd: # end of file or end of plugin definition
                            break # stop the loop
                        t = t[4:].strip() # strip the "### " from the start of the line
                        try:
                            n, v = t.split("=", 1) # name-value pair
                            if n and v and n!=PSource: # don't allow plugin authors to overwrite the Source
                                plugin[n] = v
                        except:
                            pass
                    # plugin is now read, check if it is correct (not all fields are really required)
                    if (plugin[PShortName]==PUnknown or 
                      plugin[PHint]==PUnknown or plugin[PClass]==PUnknown):
                        if DEBUGMODE:
                            print "ERROR reading plugin from file\n   %s\n(one or more required fields are unknown)" % filename
                            print "ShortName:", plugin[PShortName]
                            print "Hint:", plugin[PHint]
                            print "Class:", plugin[PClass]
                        plugin = {} # invalid, so reset to nil
                    if plugin: # add to plugins list if it's a valid plugin
                        self.plugs.append(plugin)
                elif t==PStartCollection: # found a plugin collection definition
                    # define empty plugin collection to work on
                    plugincoll = {PLongName: PUnknown,
                                  PShortName: PUnknown,
                                  PHint: PUnknown,
                                  PVersion: PUnknown,
                                  PAuthor: PUnknown,
                                  PSource: filename,
                                  PPlugins: [],
                                 }
                    emptyplug = {PShortName: PUnknown,
                                 PHint: PUnknown,
                                 PClass: PUnknown,
                                 PSource: filename,
                                     }
                    plugs = {} # is used to track plugins temporarily, before they're put in plugincoll[PPlugins]
                    while 1: # read collection
                        t = pf.readline().strip()
                        if (not t) or (t==PEnd): # end of file or end of plugin collection definition
                            break # stop reading
                        t = t[4:].strip()
                        try:
                            n, v = t.split("=", 1) # name-value pair
                        except:
                            n, v = None, None
                        if n and v and n!=PSource: # don't allow plugin authors to overwrite the Source
                            # Determine whether we're reading the collection properties or the
                            # properties of an individual plugin. If the latter is the case,
                            # names will have the form <number>.<property>, with the property
                            # being the actual name and the number being the number of the plugin.
                            # Actually the <number> is ignored and can have any value (even strings), 
                            # as long as all properties of the same plugin are prepended by the same 
                            # <number>.
                            if n.split(".", 1)[0]==n: # if this is not the case, we've started the individual plugin definitions
                                plugincoll[n] = v
                            else: # currently reading the properties of one plugin
                                i, n = n.split(".", 1)
                                if not plugs.has_key(i):
                                    plugs[i] = copy.deepcopy(emptyplug)
                                plugs[i][n] = v
                    # reading done, now process the data
                    keys = plugs.keys()
                    keys.sort()
                    for key in keys: # for every plugin in the collection
                        plug = plugs[key]
                        if (plug[PShortName]==PUnknown or plug[PHint]==PUnknown 
                          or plug[PClass]==PUnknown):
                            if DEBUGMODE:
                                print "ERROR reading plugin from file\n   %s\n(one or more required fields are unknown)" % filename
                                print "ShortName:", plug[PShortName]
                                print "Hint:", plug[PHint]
                                print "Class:", plug[PClass]
                            plug = {}
                        if plug: # is empty if something was wrong
                            plugincoll[PPlugins].append(plug)
                    if (plugincoll[PShortName]==PUnknown or not plugincoll[PPlugins]):
                        # PShortName is required (PHint NOT! in this case because it's not even
                        # shown (at least not under Windows); there must be PPlugins in there 
                        # otherwise there's no point to having a PluginCollection
                        if DEBUGMODE:
                            print "ERROR reading plugin collection from file\n   %s\n(one or more required fields are unknown)" % filename
                            print "ShortName:", plugincoll[PShortName]
                            print "Hint:", plugincoll[PHint]
                            print "Number of plugins:", len(plugincoll[PPlugins])
                        plugincoll = {}
                    if plugincoll: # if there is a valid plugin collection
                        self.plugcolls.append(plugincoll)
                    
        finally:
            pf.close();
        
    def doSize(self, event):
        """Make sure window doesn't become too small."""
        # resize window
        width, height = self.GetSize()
        self.SetSize((max(width, engine.getSetting("WindowWidthMin", 640)), 
                      max(height, engine.getSetting("WindowHeightMin", 480))))
        self.Layout()
        
    def loadSettings(self):
        """Gets GUI-related settings and applies them."""
        # set window size and position; make sure it's on screen
        # and has a reasonable size
        system = wx.wxSystemSettings_GetMetric
        pos = list(engine.getSetting(WIN_POS, (-1,-1)))
        if pos[0]<-1: pos[0] = 0
        if pos[1]<-1: pos[1] = 0
        if pos[0]>system(wx.wxSYS_SCREEN_X)-50:
            pos[0] = 0
        if pos[1]>system(wx.wxSYS_SCREEN_Y)-50:
            pos[1] = 0
        self.SetPosition(pos)
        size = list(engine.getSetting(WIN_SIZE, (-1, -1)))
        if size[0]>system(wx.wxSYS_SCREEN_X):
            size[0] = 640
        if size[1]>system(wx.wxSYS_SCREEN_Y):
            size[1] = 480
        size[0] = max(size[0], 640)
        size[1] = max(size[1], 480)
        self.SetSize(size)
        
        # set splitters
        asize = engine.getSetting(FEEDSLIST_WIDTH, self.MainSplitter.GetSashPosition())
        self.MainSplitter.SetSashPosition(max(asize, 20))
        asize = engine.getSetting(TOPICSLIST_HEIGHT, self.RightSplitter.GetSashPosition())
        self.RightSplitter.SetSashPosition(max(asize, 20))
        
        # set the column widths
        colwidths = [110, 45, 20, 20, 20] # last column widths are just dummies
        for i in range(0, self.FeedsList.GetColumnCount()):
            self.FeedsList.SetColumnWidth(i, engine.getSetting(COLUMNS+FEEDS+str(i), colwidths[i]))
        for i in range(0, self.FeedTopics.GetColumnCount()):
            self.FeedTopics.SetColumnWidth(i, engine.getSetting(COLUMNS+TOPICS+str(i), 300))
        
        # get other settings
        global MILLISECONDSREAD
        MILLISECONDSREAD = engine.getSetting(MS_READ, 250)
        self.updateOptionsMenu()
        self.redrawAll()
        
    def updateOptionsMenu(self):
        """Updates the way the options menu looks, applying the
           settings from the engine where necessary."""
        self.MenuBar.Check(pearsgui.ID_REMOVEIMAGES, 
                           engine.getSetting(pearsengine.TOPICTEXT_REMOVE_IMG, True))
        self.MenuBar.Check(pearsgui.ID_PARANOIDSAVING, 
                           engine.getSetting(pearsengine.PARANOIDSAVING, False))
        self.MenuBar.Check(pearsgui.ID_STORECOLUMNS, 
                           engine.getSetting(pearsengine.STORECOLUMNS, True))
        self.MenuBar.Check(pearsgui.ID_REVERSEORDER, 
                           engine.getSetting(pearsengine.REVERSEORDER, False))
        self.MenuBar.Check(pearsgui.ID_SHOWDIFF, 
                           engine.getSetting(pearsengine.SHOWDIFF, True))
        self.MenuBar.Check(pearsgui.ID_SMARTDISPLAY, 
                           engine.getSetting(pearsengine.SMARTDISPLAY, True))
        self.MenuBar.Check(pearsgui.ID_PLUGINWARN,
                           engine.getSetting(pearsengine.PLUGINWARN, True))
        tb = engine.getSetting(PEARSRIGHTCLICKSTYLE, True)
        self.MenuBar.Check(pearsgui.ID_RIGHTCLICKPEARS, tb)
        self.MenuBar.Check(pearsgui.ID_RIGHTCLICKSYSTEM, not tb)
        self.saveSettings() # do this, because otherwise some settings will be blank
                            # this is an extra saveguard in case I forget to specify 
                            # a default parameter in one of the getSetting() calls
        
    def doExit(self, event):
        """Shut down the program"""
        self.Close(True)
        
    def launchShell(self, event):
        """Launch PyCrust"""
        try: #only launch if not already exist
            self.shell.Show()
        except:
            from wxPython.lib.PyCrust import shell, filling
            self.shell = wx.wxFrame(self, -1, "Pears inspector")
            self.shell.SetIcon(self.GetIcon())
            #win = wx.wxSplitterWindow(self.shell, -1, size=(640, 480))
            #shellWin = shell.Shell(win, -1, introText="Direct access to Pears.\nNote: you might break things!")
            fillingWin = filling.Filling(self.shell, -1, size=(640, 480),
                                         rootObject=globals(),
                                         rootIsNamespace=1
                                         )
            #win.SplitHorizontally(shellWin, fillingWin)
            self.shell.Show()
        
    def uptimeToStr(self, uptime):
        """Converts the uptime to a string (days, hours, minutes and seconds)."""
        secinday = 24*60*60
        secinhour = 60*60
        secinmin = 60
        days = uptime // secinday
        uptime = uptime % secinday
        hours = uptime // secinhour
        uptime = uptime % secinhour
        mins = uptime // secinmin
        secs = (uptime % secinmin)//1
        template = "%02d:%02d:%02d"
        if days:
            template = "%d day(s) and "+template
            return template % (days, hours, mins, secs)
        else:
            return template % (hours, mins, secs)
            
    def getSessionUptime(self):
        """Returns the uptime for the current session"""
        now = time.time()
        if now-STARTTIME>0: # avoid problems if system clock set back in the mean time
            return now-STARTTIME
        else:
            return 0
        
    def getTotalUptime(self):
        """Calculates the total uptime of the program."""
        return engine.getSetting(TOTALUPTIME, 0) + self.getSessionUptime()
        
    def saveSettings(self):
        """Saves all GUI-related settings."""
        # save settings
        engine.setSetting(WIN_POS, self.GetPositionTuple())
        engine.setSetting(WIN_SIZE, self.GetSizeTuple())
        engine.setSetting(FEEDSLIST_WIDTH, self.MainSplitter.GetSashPosition())
        engine.setSetting(TOPICSLIST_HEIGHT, self.RightSplitter.GetSashPosition())
        
        for i in range(0, self.FeedsList.GetColumnCount()):
            engine.setSetting(COLUMNS+FEEDS+str(i), self.FeedsList.GetColumnWidth(i))
        for i in range(0, self.FeedTopics.GetColumnCount()):
            engine.setSetting(COLUMNS+TOPICS+str(i), self.FeedTopics.GetColumnWidth(i))
        
        engine.setSetting(MS_READ, MILLISECONDSREAD)
        engine.setSetting(PLAINDRAWMODE, False)
        
        engine.setSetting(TOTALUPTIME, self.getTotalUptime())
        
    def doClose(self, event):
        """Handler of wx.EVT_CLOSE"""
        self.saveSettings()
        
        # shut down engine
        engine.close()
        
        # stop the RADIOPORT listener
        self.RPCOnClose(event)        
        
        # shut down program
        self.Destroy()
        
    def setRemoveImages(self, event):
        """Toggles the loading of images on and off"""
        currentstatus = engine.getSetting(pearsengine.TOPICTEXT_REMOVE_IMG, True)
        engine.setSetting(pearsengine.TOPICTEXT_REMOVE_IMG, not currentstatus)
        
    def setParanoidSaving(self, event):
        """Toggles the paranoid saving method on and off"""
        currentstatus = engine.getSetting(pearsengine.PARANOIDSAVING, False)
        engine.setSetting(pearsengine.PARANOIDSAVING, not currentstatus)
        
    def setShowDiff(self, event):
        """Toggles the diff showing in the topic view on and off"""
        currentstatus = engine.getSetting(pearsengine.SHOWDIFF, True)
        engine.setSetting(pearsengine.SHOWDIFF, not currentstatus)
        
    def setSmartDisplay(self, event):
        """Toggles the smart fields display (no duplicates) on and off"""
        currentstatus = engine.getSetting(pearsengine.SMARTDISPLAY, True)
        engine.setSetting(pearsengine.SMARTDISPLAY, not currentstatus)
        
    def setPluginWarn(self, event):
        """Toggles the warning message before a plugin is executed on and off"""
        currentstatus = engine.getSetting(pearsengine.PLUGINWARN, True)
        engine.setSetting(pearsengine.PLUGINWARN, not currentstatus)
        
    def setStoreColumns(self, event):
        """Toggles the storing of column width on and off"""
        currentstatus = engine.getSetting(STORECOLUMNS, False)
        engine.setSetting(STORECOLUMNS, not currentstatus)
        
    def setRightclick(self, event):
        """Toggles right-click behavior between Pears style and
           system style (popup menu)."""
        id = event.GetId()
        if id==pearsgui.ID_RIGHTCLICKPEARS:
            engine.setSetting(PEARSRIGHTCLICKSTYLE, True)
        else:
            engine.setSetting(PEARSRIGHTCLICKSTYLE, False)
        
    def redrawAll(self):
        """Redraws all controls on the screen."""
        #selfeed = self.FeedsList.CurrentItem
        #seltopic = self.TopicsList.CurrentItem
        for feednr in range(self.FeedsList.GetItemCount()):
            self.setReadStatusOfFeed(feednr)
        if type(self.FeedsList.CurrentItem)==type(5):
            self.displayItems(self.FeedsList.CurrentItem)
        # set font for HTML
        sf = self.GetFont()
        SetHTMLFonts(self.TopicText, sf)
        #if validIds():
        #    pass
        
    def setFont(self, event):
        """Changes the font face used in the list controls"""
        data = wx.wxFontData()
        data.SetAllowSymbols(True) # allow symbol fonts in Windows (for weird alphabets)
        data.EnableEffects(False) # don't show the font markup button in Windows
        data.SetRange(8, 20) # allowed font sizes in Windows
        data.SetShowHelp(False) # don't show help
        data.SetInitialFont(self.GetFont())
        dlg = wx.wxFontDialog(self, data)
        if dlg.ShowModal() == wx.wxID_OK:
            data = dlg.GetFontData()
            font = data.GetChosenFont()
            engine.setSetting(FONT, {FFACE: font.GetFaceName(),
                                     FSIZE: font.GetPointSize(),
                                     FFAMILY: font.GetFamily()})
            # redraw the screen
            self.redrawAll()
        dlg.Destroy()
        
    def setBrowser(self, event):
        """Changes the browser used to open URLs"""
        dlg = BrowserDialog(parent=self, id=-1, title="Browser configuration")
        if dlg.ShowModal()==1:
            pass
        dlg.Destroy()
        
    def setReverseOrder(self, event):
        """Toggles the storing of reverse order on and off"""
        currentstatus = engine.getSetting(REVERSEORDER, False)
        engine.setSetting(REVERSEORDER, not currentstatus)
        
    def setTimeRead(self, event):
        """Allows the user to specify after how many seconds of being
           open a topic should be regarded as read."""
        global MILLISECONDSREAD
        currenttime = "%.2f" % (MILLISECONDSREAD/1000.0)
        if MILLISECONDSREAD > 3600:
            currenttime = "never"
        asis = "keep current value"
        never = "never"
        choices = [asis+" (%s)" % currenttime, 
                   "0.00", "0.25", "0.50", "0.75", "1.00",
                   "1.25", "1.50", "1.75", "2.0", "2.5",
                   "3.0", "3.5", "4.0", "5.0", "10", "15", "20", 
                   "30", "60", never]
        dlg = wx.wxSingleChoiceDialog(self, 
                                   'Choose after how many seconds\na topic should be automatically\nmarked as read.', 
                                   'Mark as read after...',
                                   choices, wx.wxOK|wx.wxCANCEL)
        if dlg.ShowModal() == wx.wxID_OK:
            sel = dlg.GetStringSelection()
            if sel.find(asis)<0: # if asis is found, no change
                if sel==never:
                    MILLISECONDSREAD = 999999999
                else:
                    try:
                        temp = int(round(1000*float(sel)))
                        MILLISECONDSREAD = temp
                    except: pass
                
        dlg.Destroy()
        
    def showHelp(self, event):
        """Show help"""
        try:
            import webbrowser
            webbrowser.open("file://"+os.path.join(PEARSLOCATION, "docs/index.html"), 1)
        except:
            dlg = wxScrolledMessageDialog(parent=self, 
                    caption="Help",
                    msg="""PEARS - alternative help (see the docs directory in your Pears folder for more elaborate information)
                    
    Pears is a relatively simple three-pane news feed aggregator (RSS/RDF reader).
        
    The left pane contains the list of feeds. You can add feeds using the Feeds menu. Right-click on a feed to update, single-click to view cached contents.
        
    The right pane contains a list of the topics available in the feed. Single-click to open the contents of a topic in the viewer window at the bottom. Right-click to toggle read/unread status.
        
    Everything is cached in Python structures stored as plain text in your Home directory (in a subdirectory called '.pears'). Use the About screen to see exactly where that is on your system.
    """)
            dlg.ShowModal()
            dlg.Destroy()
        
    def showAbout(self, event):
        """Show about screen"""
        dlg = AboutDialog(parent=self, id=-1, title="About Pears")
        dlg.ShowModal()
        dlg.Destroy()
        
    def moveFeed(self, event):
        """Moves a feed one up or down (depending on which menu item
           called the method."""
        if not self.validIds(checkTopicsId=False): # feed must be selected
            return None
        try:
            if event.GetId()==pearsgui.ID_MOVEUP: diff = -1 # move up
            elif event.GetId()==pearsgui.ID_MOVEDOWN: diff = +1 # move down
            else: diff = 0
            if DEBUGMODE: print "Moving feed %s by %s" % (self.FeedsList.CurrentItem, diff)
        except:
            return None # invalid call
        oldpos = self.FeedsList.CurrentItem
        self.storeStatus()
        newpos = engine.moveFeed(self.FeedsList.CurrentItem, self.FeedsList.CurrentItem+diff)
        if oldpos<>newpos: # update feeds list
            listitem = self.FeedsList.GetItem(oldpos)
            self.FeedsList.DeleteItem(oldpos)
            listitem.SetId(newpos)
            self.FeedsList.InsertItem(listitem)
            self.stored.CurrentFeed = newpos
            self.restoreStatus()
        self.FeedsList.EnsureVisible(newpos)
            
    def sortFeedsColumn(self, event):
        """Sorts the feeds when the user clicks on a column header. Depending
           on which column was clicked, sorting is either done alphabetically,
           or by the number of new topics."""
        # column number might be -1 if header clicked, but not on a column
        print event.GetColumn()
        oldorder = u''.join(engine.getFeedsNames()) # remember current order
        if event.GetColumn()==0: # sort by title
            # if the list is already sorted, then reverse
            engine.sortFeeds(inverse=False)
        elif event.GetColumn()==1: # sort by unread
            engine.sortFeedsUnread()
        if oldorder==u''.join(engine.getFeedsNames()):
            # nothing changed, was already sorted: inverse
            engine.reverseFeeds()
        if event.GetColumn()>-1:
            self.showFeeds()
            
    def sortFeeds(self, event):
        """Sorts the feeds in a certain order, depending on which menu item
           called the method."""
        try:
            id = event.GetId()
        except:
            return None # invalid call
            
        #~ t = 'You are about to change the order of your feeds list.\n' + \
            #~ 'This action can not be undone. Proceed?'
        #~ if not self.askConfirmation(amessage=t, acaption="Sort feeds?"):
            #~ return None
            
        self.FeedTopics.DeleteAllItems()
        self.FeedTopics.CurrentItem = None
        self.TopicText.SetPage("")
        
        if id==pearsgui.ID_SORTFEEDSALFA: 
            engine.sortFeeds(inverse=False)
        elif id==pearsgui.ID_SORTFEEDSINVALFA: 
            engine.sortFeeds(inverse=True)
        elif id==pearsgui.ID_SORTFEEDSREVERSE:
            engine.reverseFeeds()
        elif id==pearsgui.ID_SORTFEEDSRANDOM:
            engine.randomizeFeeds()
        elif id==pearsgui.ID_SORTFEEDSUNREAD:
            engine.sortFeedsUnread()
            
        self.showFeeds()        
        
    def getFileName(self, msg="Choose a file", defdir="", 
                    deffile="", wild=(("Any file","*.*")), 
                    stl=wx.wxSAVE|wx.wxOVERWRITE_PROMPT):
        """Gets and returns a file name and the chosen filter type. 
           If cancelled, returns the tuple (None, None)."""
        # translate tuple wild to a string as used by the file dialog
        if type(wild) not in [type("s"), type(u"s")]:
            wild = [ "|".join(("%s (%s)" % (items[0].strip(), items[1].strip()), \
                                "%s" % items[1].strip())) \
                     for items in wild ]
            wild = "|".join(wild)
        dlg = wx.wxFileDialog(self, message=msg, defaultDir=defdir,
                              defaultFile=deffile, wildcard=wild, 
                              style=stl)
        if dlg.ShowModal()==wx.wxID_CANCEL:
            res = None, None
        else:
            res = dlg.GetPath(), dlg.GetFilterIndex()
        dlg.Destroy()
        return res
        
    def importFeeds(self, event):
        """Import feeds from a XML file."""
        # first get the file which should be imported
        fn = self.getFileName(msg="Import...", 
                              wild=[("OPML files","*.opml")],
                              stl=wx.wxOPEN)
        fn = fn[0] # fn by default is a tuple of filename and selected filter
        if fn: # if filename specified
            feeds = engine.importOpml(fn)
            if feeds: # if there are any new feeds
                dlg = ImpexDialog(self, id=-1, 
                                  title="Select feeds to import from %s" % os.path.basename(fn))
                dlg.setOkButtonLabel("Import selection")
                dlg.setItems(feeds)
                if dlg.ShowModal()==wx.wxID_OK:
                    feeds = dlg.getSelection()
                    # now add the feeds to the engine
                    for feed in feeds:
                        engine.addFeed(location=feed[0], title=feed[1],
                                       channel={"tagline": feed[2]})
                        feedindex = self.FeedsList.GetItemCount()
                        self.FeedsList.InsertStringItem(feedindex, dispify(engine.getFeedsNames()[-1]))
                        self.setReadStatusOfFeed(feedindex)
                dlg.Destroy()
                self.resizeFeedsList()
            else:
                self.askConfirmation(amessage="The file (%s) was invalid or did not contain any new feeds"  % fn,
                                     acaption="Import",
                                     astyle=wx.wxOK|wx.wxICON_INFORMATION)
        
    def exportFeeds(self, event):
        """Exports feeds to a XML file."""
        FILTER = "title"
        FUNCTION = "function"
        # map available export options to engine functions; keys MUST
        # be consecutive integers starting with 0
        possibilities = {0: {FILTER: ("OPML file", "*.opml"), 
                             FUNCTION: engine.exportOpml}}
        # build the filter to be used in the open dialog
        filter = []
        keys =  possibilities.keys()
        keys.sort()
        for key in keys:
            filter.append(possibilities[key][FILTER])
        fn, action = self.getFileName(msg="Export...", 
                                      defdir=engine.savedir,
                                      deffile="pears.opml", 
                                      wild=filter,
                                      stl=wx.wxSAVE|wx.wxOVERWRITE_PROMPT)
        if not fn: # if no filename was specified, stop
            return
        # select the feeds to export
        dlg = ImpexDialog(self, id=-1, 
                          title="Select feeds to export to %s" % os.path.basename(fn))
        dlg.setOkButtonLabel("Export selection")
        dlg.setItems(engine.getFeedsForExport())
        dlg.selectAll()
        if dlg.ShowModal()==wx.wxID_OK:
            feeds = dlg.getSelection()
            if feeds: # if any selected
                # now export the feeds according to the selected file type
                possibilities[action][FUNCTION](fn, feeds)
        dlg.Destroy()
        
    def showFeedProperties(self, event):
        """Show (and modify) feed properties"""
        if not self.validIds(checkTopicsId=False):
            return None
        dlg = AddDialog(parent=self, id=-1, title="Properties")
        tempfeedprops = engine.getFeed(self.FeedsList.CurrentItem)
        dlg.FeedLocation.SetValue(tempfeedprops[LOCATION])
        dlg.FeedTitle.SetValue(dispify(tempfeedprops[TITLE]))
        dlg.TopicsStore.SetStringSelection(str(tempfeedprops[CACHESIZE]))
        dlg.SetStatus("")
            
        # determine update interval
        up = tempfeedprops[UPDATEINTERVAL]
        if up>=NEVER:
            up = "Never"
        else:
            # convert to hours if possible; don't pay attention to 10 minutes or less inaccuracy
            mask = "%s hours"
            if (up%60<=AUTOUPDRESOLUTION or up%60>=60-AUTOUPDRESOLUTION) and up>=60: 
                # five minutes inaccuracy ignored, rouding appropriately
                up = up//60
            else: # specify in minutes
                up = max(up, 1) # at least 30 minutes update interval
                mask = "%s minutes" 
            up = mask % up
        dlg.AutoRefresh.Append(up + " [current]")
        dlg.AutoRefresh.SetSelection(dlg.AutoRefresh.GetCount()-1)
            
        dlg.feedprops = tempfeedprops[CHANNEL]
        dlg.showProps()
        dlg.prepareForAdd()
        dlg.AddButton.SetLabel("Modify")
        dlg.AddButton.SetToolTipString("Apply modifications")
        dlg.FeedTitle.SetToolTipString("This title is displayed in the feeds list")
        dlg.FeedLocation.SetToolTipString("URL of selected feed")
        dlg.SetTitle("Modify feed")
        if dlg.ShowModal():
            index = self.FeedsList.CurrentItem
            self.FeedsList.SetStringItem(index,
                                         0, 
                                         engine.getFeedsNames()[index])
            self.setReadStatusOfFeed(index)
            self.displayItems(index)
        dlg.Destroy()
        
    # DO NOT CHANGE THE NAME OF THIS METHOD WITHOUT UPDATING THE
    # OnData method of the MyURLDropTarget class in pearsgui
    def addFeedRequest(self, *args, **kwargs):
        """Handle a localhost HTTP 'CoffeeCup' request to add subscription"""
        if DEBUGMODE: print 'addFeedRequest', `(args, kwargs)`
        if 'url' in kwargs:
            url = kwargs['url'].strip()
            if (url=="") or (not url.startswith('http://')):
                return "Invalid URL for feed: cannot add: "+url
            elif (url in engine.getFeedsLocations()):
                return "Already subscribed to feed: "+url
            else:
                self.Iconize(True)
                self.Iconize(False)
                self.Show(True)
                self.Raise()
                Pears.Yield()
                self.addFeed(None, url)
                if (url in engine.getFeedsLocations()):
                    return "Added feed: "+url
                else:
                    return "Did not add feed: "+url
        
    def addFeed(self, event, url=""):
        """Add a feed"""
        if DEBUGMODE: print "addFeed(url=%s)" % url
        dlg = AddDialog(parent=self, id=-1, title="Add feed")
        dlg.FeedLocation.SetValue(url)
        dlg.doEnable(None) # just in order to set the status
        if dlg.ShowModal()==1:
            feedindex = self.FeedsList.GetItemCount()
            self.FeedsList.InsertStringItem(feedindex, dispify(engine.getFeedsNames()[-1]))
            self.setReadStatusOfFeed(feedindex)
        dlg.Destroy()
        self.resizeFeedsList()
            
    def enableAll(self, enabled=True):
        """Enable all controls."""
        self.Enable(enabled)
        #~ self.FeedsList.Enable(enabled)
        #~ self.FeedTopics.Enable(enabled)
        #~ self.TopicText.Enable(enabled)
        #~ for menuitem in range(self.MenuBar.GetMenuCount()):
            #~ self.MenuBar.EnableTop(menuitem, enabled)
        try: #once got a weird error about this, so no try-except
            Pears.Yield()
        except:
            pass
        
    def disableAll(self):
        """Disable all controls, wrapping enableAll()"""
        self.enableAll(enabled=False)
        
    def askConfirmation(self, amessage="Are you sure?", acaption="Confirm", 
                        astyle=wx.wxYES_NO|wx.wxICON_QUESTION|wx.wxNO_DEFAULT):
        """Asks for confirmation to a certain question.
           Returns True if the answer is "Yes", False otherwise."""
        dlg = wx.wxMessageDialog(parent=self, message=amessage, 
                                 caption=acaption, style=astyle)
        res = dlg.ShowModal()
        dlg.Destroy()
        if res in [wx.wxID_YES, wx.wxID_OK]:
            return True
        else:
            return False
        
    def removeFeed(self, event):
        """Remove selected feed"""
        if not self.validIds(checkTopicsId=False):
            return None
        self.disableAll()
        try:
            temptext = 'You are about to delete feed \"%s\".\n' + \
                       'This action can not be undone. Proceed?'
            temptext = temptext % self.FeedsList.GetItem(self.FeedsList.CurrentItem).GetText()
            # ask confirmation
            ok = self.askConfirmation(amessage=temptext, acaption="Delete feed?", 
                                      astyle=wx.wxYES_NO|wx.wxICON_QUESTION|wx.wxNO_DEFAULT)
            if ok:
                if engine.removeFeed(self.FeedsList.CurrentItem):
                    self.FeedsList.DeleteItem(self.FeedsList.CurrentItem)
                    self.FeedTopics.DeleteAllItems()
                    self.FeedTopics.CurrentItem = None
                    self.FeedsList.CurrentItem = None
                    self.TopicText.SetPage("")
        finally:
            self.resizeFeedsList()
            self.enableAll()
            
    def drawItem(self, list, itemnumber, 
                 setToUnread=True, 
                 setToCacheExceeded=False, 
                 setToUpdating=False):
        """Draws an item."""
        currentitem = list.GetItem(itemnumber)
        if engine.getSetting(PLAINDRAWMODE): #do not use colors and stuff
            def addOrRemoveMarker(marker, doadd):
                """Adds or removes a marker from an item."""
                txt = currentitem.GetText()
                if doadd:
                    if txt.find(marker)<0:
                        currentitem.SetText(marker+txt)
                else:
                    if txt.find(marker)==0:
                        currentitem.SetText(txt[len(marker):])
                
            addOrRemoveMarker("* ", setToUnread)
        else: #advanced drawing
            #can't modify existing font, due to problems with wxGTK, so make new one
            sf = self.GetFont()
            savedfont = engine.getSetting(FONT, {FFACE: sf.GetFaceName(),
                                                 FSIZE: sf.GetPointSize(),
                                                 FFAMILY: sf.GetFamily()})
            newfont = wx.wxFont(pointSize=savedfont[FSIZE], 
                             #family=wx.wxDEFAULT, style=wx.wxNORMAL, weight=wx.wxNORMAL)
                                family=savedfont[FFAMILY],
                                style=wx.wxNORMAL, 
                                weight=wx.wxNORMAL)
            newfont.SetFaceName(savedfont[FFACE])
            if setToUnread:
                weight = wx.wxBOLD
            else:
                weight = wx.wxNORMAL
            newfont.SetWeight(weight)
            
            #~ if setToCacheExceeded:
                #~ style = wxITALIC
            #~ else:
                #~ style = wxNORMAL
            #~ newfont.SetStyle(style)
            #~newfont.SetStyle(style)
            
            newfont.SetUnderlined(setToUpdating)
            
            currentitem.SetFont(newfont)
            
            if setToCacheExceeded:
                currentitem.SetTextColour(wx.wxNamedColour("forest green"))
            else:
                currentitem.SetTextColour(wx.wxNamedColour("black"))
        
        list.SetItem(currentitem)
        
    def displayItems(self, feednumber):
        """Display the items of feednumber."""
        if self.LockedFeed != None:
            engine.unlockFeed(self.LockedFeed)
        self.LockedFeed = engine.lockFeed(feednumber)        
        self.FeedTopics.DeleteAllItems()
        self.FeedTopics.CurrentItem = None
        self.TopicText.SetPage("")
        topics = engine.getTopics(feednumber)
        self.setReadStatusOfFeed(feednumber)
        counter = 0
        for item in topics:
            x = self.FeedTopics.GetItemCount()
            self.FeedTopics.InsertStringItem(x, dispify(item[0]))
            self.drawItem(self.FeedTopics, x, 
                          setToUnread=item[1],
                          setToCacheExceeded=counter>=engine.getFeedCacheSize(self.FeedsList.CurrentItem))
            counter += 1 
        self.resizeFeedTopics()
        
    def showFeedItems(self, event):
        """Get the items in the selected feed and display them
           in FeedTopics. Also wraps self.displayItems()"""
        if self.ReadTimer.IsRunning():
            self.ReadTimer.Stop()
        try: # assign CurrentItem if clicked on list
            self.FeedsList.CurrentItem = event.m_itemIndex
        except:
            pass
        self.displayItems(self.FeedsList.CurrentItem)
        if engine.getSetting(pearsengine.PARANOIDSAVING, False):
            engine.saveAll()
            
    class empty(object):
        """Dummy class, used to store properties by storeStatus."""
        pass
            
    def storeStatus(self):
        """Stores selected items and the likes. Useful in combination with
           self.restoreStatus()"""
        self.stored = self.empty()
        ss = self.stored
        ss.CurrentFeed = self.FeedsList.CurrentItem
        ss.CurrentTopic = self.FeedTopics.CurrentItem
        ss.FeedsCount = self.FeedsList.GetItemCount()
        ss.TopicsCount = self.FeedTopics.GetItemCount()
        if ss.CurrentFeed!=None:
            self.CurrentFeedText = self.FeedsList.GetItemText(ss.CurrentFeed)
        else:
            self.CurrentFeedText = ""
        if ss.CurrentTopic!=None:
            ss.CurrentTopicText = self.FeedTopics.GetItemText(ss.CurrentTopic)
        else:
            self.CurrentTopicText = ""
        ss.Focus = wx.wxWindow_FindFocus()
        
    def restoreStatus(self):
        """Restores selected items and the likes after an update."""
        try:
            ss = self.stored
        except: # self.stored doesn't exist
            return
        # restore feed selection
        if type(ss.CurrentFeed)==type(4) and \
           0 <= ss.CurrentFeed < self.FeedsList.GetItemCount():
            self.FeedsList.CurrentItem = ss.CurrentFeed
            self.FeedsList.SetItemState(ss.CurrentFeed, wx.wxLIST_STATE_SELECTED, wx.wxLIST_STATE_SELECTED)
        # restore topic selection
        # first find out if there are new topics in there
        newtopics = self.FeedTopics.GetItemCount() - ss.TopicsCount
        # restore if there are 0 or more new topics
        # and if at the old position the same item is present
        # this means that if that item was updated (changed),
        # selection is not restored
        # This method is not 100% error-free, but it will work in
        # 99.999% of the cases.
        if newtopics>=0 and \
           ss.CurrentTopic!=None and \
           self.FeedTopics.GetItemText(ss.CurrentTopic + newtopics)==ss.CurrentTopicText:
            self.FeedTopics.CurrentItem = ss.CurrentTopic
            self.FeedTopics.SetItemState(ss.CurrentTopic+newtopics, wx.wxLIST_STATE_SELECTED, wx.wxLIST_STATE_SELECTED)
            self.ReadTimer.Stop()
        try:
            ss.Focus.SetFocus()
        except:
            pass
            
    def addMenuItem(self, menu, caption, hint, eventhandler):
        """Adds a menu item with a caption and hint to the menu. Also binds
           the event handler."""
        m_id = pearsgui.ni()
        menu.Append(m_id, caption, hint)
        wx.EVT_MENU(self, m_id, eventhandler)
            
    def rclickFeed(self, event):
        """Decides what to do when a feed was right-clicked:
           either update, or show popup menu, depending on setting."""
        self.FeedsList.CurrentItem = event.m_itemIndex # if called by mouse message
        self.FeedsList.SetItemState(self.FeedsList.CurrentItem, wx.wxLIST_STATE_SELECTED, wx.wxLIST_STATE_SELECTED)
        if engine.getSetting(PEARSRIGHTCLICKSTYLE, True):
            self.updateFeed(event)
        else: # show popup menu if not Pears-style right-click behavior
            # make popup menu
            mn = wx.wxMenu()
            self.addMenuItem(mn, '&Update', 'Update this feed', self.updateFeed)
            self.addMenuItem(mn, 'Mark as read', 'Mark all topics in this feed as read', self.markAllTopicsRead)
            self.addMenuItem(mn, 'Mark as unread', 'Mark all topics in this feed as unread', self.markAllTopicsUnread)
            self.addMenuItem(mn, '&Empty cache', 'Remove all cached topics in this feed', self.deleteAllTopics)
            mn.AppendSeparator()
            self.addMenuItem(mn, '&Properties', 'Show properties of this feed', self.showFeedProperties)
            self.addMenuItem(mn, '&Remove', 'Remove this feed', self.removeFeed)
            # show menu
            self.FeedsList.PopupMenu(mn, event.GetPoint())
            # clean up
            for mi in mn.GetMenuItems():
                mn.Remove(mi.GetId())
                mi.Destroy()
            mn.Destroy()
        
    def updateFeed(self, event):
        """Update the selected feed."""
        if not self.validIds(checkTopicsId=False):
            return None
        self.disableAll()
        self.storeStatus()
        self.setReadStatusOfFeed(self.FeedsList.CurrentItem, isBeingUpdated=True)
        try:
            Pears.Yield()
            engine.updateFeed(self.FeedsList.CurrentItem)
            self.showFeedItems(event)
        finally:
            self.setReadStatusOfFeed(self.FeedsList.CurrentItem)
            self.enableAll()
            self.restoreStatus()
        
    def updateAll(self, event):
        """Update all feeds."""
        # find out if updateAll was triggered by a wxTimer event
        try:
            force = event.GetInterval() < -2 # returns False if event triggered by timer
            timercall = True
        except:
            #~force = True # means the update request was done manually
            force = event.GetId() == pearsgui.ID_UPDATEALLFORCE
            timercall = False
            
        # if triggered by timer, only auto-update if window iconized, otherwise it will
        # jump to front which is very annoying
        if timercall and not self.IsIconized():
            return None
        else:
            if DEBUGMODE: print force, self.IsIconized()
            
        # used in multi-treading capabilities:
        def _updateFeed(feednumber):
            engine.updateFeed(feednumber, force)
            return feednumber
        # end of _updateFeed()
            
        futures = [] # a list of futures (functions running in their own threads)
        toupdate = [] # a list of feeds which must be updated
            
        self.disableAll()
        self.storeStatus()
        try:
            Pears.Yield()
        except:
            pass
        try:
            for feednumber in range(self.FeedsList.GetItemCount()):
                if not force and not engine.autoupdateRequired(feednumber):
                    # if no forced update and not time for an update, then go to next item
                    continue
                toupdate.append(feednumber)
            # now we know which feeds must be updated, so deploy the futures
            while toupdate or futures: 
            # while there are still feeds in need of updating or in process of being updated
                # if necessary and possible, add more feeds to the futures
                while len(futures)<SIMULTANEOUSUPDATE and toupdate:
                    # move a feed number to update
                    feednumber = toupdate.pop(0)
                    print "Updating", feednumber
                    futures.append(future.Future(_updateFeed, feednumber))
                    # show it as updating
                    self.setReadStatusOfFeed(feednumber, isBeingUpdated=True)
                    Pears.Yield()
                # now let it run for a while, otherwise it will consume 100% CPU
                time.sleep(.2)
                # check if any of them are done and if so, display that
                for function in futures:
                    if function.isDone():
                        #feednumber = function()
                        #print "Done", feednumber
                        feednumber = function.getParams()[0] # patch by Manuel V.
                        print "Done", feednumber #,function() # patch by Manuel V.
                        futures.remove(function)
                        # show it as updated
                        try:
                            self.setReadStatusOfFeed(feednumber)
                        except:
                            print "ERROR updating feednumber:", feednumber
                        Pears.Yield()
            if self.validIds(checkTopicsId=False):
                self.displayItems(self.FeedsList.CurrentItem)
        finally:
            self.enableAll()
            self.restoreStatus()
            
    def topicAsHtml(self, topic):
        """Returns a HTML version of topic."""
        def htmlize(t):
            savedfont = engine.getSetting(FONT)
            text = []
            #text.append(u"""<font face="%s" size="%s">""" % (savedfont[FFACE], ""))
            # The code below was heavily reworked in version 1.7 beta 11
            #~ exclude = [PREVIOUS, 
                       #~ "categories", # introduced in newer feedparser versions
                       #~ ] # keeps track of keys which are to be excluded at the end
                
            title = t.get("title", "")
            summary = t.get("summary", "")
            content = engine.getContent(t)
            link = engine.getLink(t)
            id = t.get("id", "")
            if id==link and engine.getSetting(pearsengine.SMARTDISPLAY, True):
                id = "" # don't show both link and id if they're identical and SMARTDISPLAY is on
            # The categories entry of the item is a list of tuples of the form:
            # (domain, output). One of these might be None. Convert that to empty string
            # to avoid errors.
            categories = t.get("categories", [])
            for i in range(len(categories)):
                categories[i] = list(categories[i])
                if not categories[i][0]: 
                    categories[i] = categories[i][1]
                else:
                    categories[i] = ": ".join(categories[i]) # merge domain and output to a single string
            categories = [ cat for cat in categories if cat ]
            author = t.get("author", "")
            comments = t.get("comments", "")
            modified = t.get("modified", "")
            date = t.get("date", "")
            
            # Make a list of the keys which are supported (some keys returned by the parser are 
            # deleted at download time, so don't worry about those).
            supported = ["title", "summary", "content", "links", "id", 
                         "categories", "author", "comments",
                         "modified", "modified_parsed", "date", "date_parsed"]
            
            # Now put the HTML together.
            if title:
                text.append(u'<font size="6"><b>%s</b></font> ' % title)
                text.append(u"<br>")
            if author:
                text.append(u"".join([u"by <u>", author, u"</u> "]))
            if categories:
                text.append(u"".join([u"in <u>", u", ".join(categories), u"</u> "]))
            if author or categories:
                text.append(u"<br><br>")
            elif title:
                text.append(u"<br>") # need an extra blank line between title and contents
            if summary:
                if content: text.append(u'<font color="#666666">')
                text.append(summary)
                if content: text.append(u"</font>")
                text.append(u"<br><br>")
            if content:
                text.append(content)
                text.append(u"<br><br>")
            if link:
                text.append(u"".join([u"Visit ", link]))
                text.append(u"<br><br>")
            if comments:
                text.append(u"".join([u"Comments are welcome on ", comments]))
                text.append(u"<br><br>")
            if date:
                text.append(u"".join([u"Posted on ", date]))
            if modified and modified!=date:
                if date:
                    text.append(u"".join([u" and modified on ", modified]))
                else:
                    text.append(u"".join([u"Modified on ", modified]))
            if date or modified:
                text.append(u"<br><br>")
            if id:
                text.append(u"".join([u"Topic id: ", id]))
                text.append(u"<br><br>")
                
            if not engine.getSetting(pearsengine.SMARTDISPLAY, True): # is it necessary to show unsupported keys?
                text.append(u"<hr>")
                keys = t.keys()
                keys.sort()
                for key in keys:
                    if key not in supported:
                        text.append(u"<b>%s</b>: %s" % (unicode(key), unicode(t[key])))
                        text.append(u"<br><br>")
                        
            text[-1] = text[-1].strip("<br>") # remove trailing blank lines
            #text.append(u"</font>") # shut down tag that was opened at the very beginning
            res = engine.modifyText(u"".join(text))
            if DISPCODE: # unicode python has DISPCODE == None
                return res.encode(DISPCODE, "xmlcharrefreplace")
            else: 
                return res
                        
        if topic.has_key(PREVIOUS) and engine.getSetting(pearsengine.SHOWDIFF, True):
            return htmldiff.textDiff(htmlize(topic[PREVIOUS]), htmlize(topic))
        else: 
            return htmlize(topic)
        
    def topicAsText(self, topic):
        """Returns a plain text version of topic."""
        htmltext = self.topicAsHtml(topic)
        htmltext = htmltext.split("<br>")
        htmltext = [engine.htmlToPlain(t, DISPCODE) for t in htmltext]
        return u"\n".join(htmltext)
        
    def showTopic(self, event):
        """Displays the current topic of the current feed
           in the TopicText window."""
        try:
            self.FeedTopics.CurrentItem = event.m_itemIndex
        except:
            pass
        self.disableAll()
        try:
            self.ReadTimer.Start(MILLISECONDSREAD, wx.wxTIMER_ONE_SHOT)
            topic = engine.getItem(self.FeedsList.CurrentItem, self.FeedTopics.CurrentItem)
            #self.TopicText.SetPage(self.topicAsHtml(topic).encode('utf8','xmlcharrefreplace'))
            self.TopicText.SetPage(self.topicAsHtml(topic))
        finally:
            self.enableAll()
            pass
        self.FeedTopics.SetFocus()
            
    def resizeFeedsList(self):
        """Resizes the columns of the feeds list"""
        if not engine.getSetting(STORECOLUMNS):
            self.FeedsList.SetColumnWidth(0, wx.wxLIST_AUTOSIZE)
            self.FeedsList.SetColumnWidth(1, wx.wxLIST_AUTOSIZE)
        
    def resizeFeedTopics(self):
        """Resizes the columns of the topics list"""
        if not engine.getSetting(STORECOLUMNS):
            self.FeedTopics.SetColumnWidth(0, wx.wxLIST_AUTOSIZE)
        
    def setReadStatusOfFeed(self, feedid, isBeingUpdated=False):
        """Takes care of displaying read/unread items and style
           of a feed specified by feedid."""
        # get total number of items and number of new items
        new, total = engine.getTopicCounts(feedid) 
        #if DEBUGMODE: print "readstatus", `(new, total, feedid)`        
        self.FeedsList.SetStringItem(feedid, 1, "%s/%s" % (new, total))
        self.drawItem(self.FeedsList, feedid, 
                      setToUnread=new>0, 
                      setToCacheExceeded=engine.isFeedCacheExceeded(feedid),
                      setToUpdating=isBeingUpdated)
        self.resizeFeedsList()
        
    def showFeeds(self):
        """Loads the feeds from the engine and displays them."""
        self.FeedsList.DeleteAllItems()
        self.FeedsList.CurrentItem = None
        for feed in engine.getFeedsNames():
            x = self.FeedsList.GetItemCount()
            self.FeedsList.InsertStringItem(x, dispify(feed))
            self.setReadStatusOfFeed(x)
            self.resizeFeedsList()
        self.FeedsList.SetFocus()
        
    def initFeeds(self):
        """Initializes the feeds list."""
        self.FeedTopics.DeleteAllItems()
        self.FeedTopics.CurrentItem = None
        self.TopicText.SetPage("")
        self.showFeeds()
        
    def markTopicRead(self, event, setToUnread=False):
        """Marks current topic as read."""
        if not self.validIds():
            return None
        self.disableAll()
        try:
            if engine.setItemRead(self.FeedsList.CurrentItem, self.FeedTopics.CurrentItem, setToUnread):
                self.drawItem(self.FeedTopics, self.FeedTopics.CurrentItem, 
                              setToUnread,
                              setToCacheExceeded=self.FeedTopics.CurrentItem>=engine.getFeedCacheSize(self.FeedsList.CurrentItem))
        finally:
            self.setReadStatusOfFeed(self.FeedsList.CurrentItem)
            self.enableAll()
        self.resizeFeedsList()
        self.resizeFeedTopics()
        self.FeedTopics.SetFocus()
        
    def markAllTopicsRead(self, event, setToUnread=False):
        """Marks all topics in current feed as read."""
        if not self.validIds(checkTopicsId=False):
            return None
        self.storeStatus()
        self.disableAll()
        try:
            if engine.setAllItemsInFeedRead(self.FeedsList.CurrentItem, setToUnread):
                self.displayItems(self.FeedsList.CurrentItem)
        finally:
            self.setReadStatusOfFeed(self.FeedsList.CurrentItem)
            self.enableAll()
        self.resizeFeedsList()
        self.resizeFeedTopics()
        self.FeedTopics.SetFocus()
        self.restoreStatus()
        
    def markTopicUnread(self, event):
        """Marks current topic as unread, wrapping markTopicRead()."""
        self.markTopicRead(event, True)
        
    def markAllTopicsUnread(self, event):
        """Marks all topics in current feed as unread, wrapping
           markAllTopicsRead()."""
        self.markAllTopicsRead(event, True)
        
    def rclickTopic(self, event):
        """Handles right-click event for the topics list. Depending
           on what the settings are, this either toggles read/unread
           or shows a popup menu."""
        self.toggleTopicRead(event)
        if engine.getSetting(PEARSRIGHTCLICKSTYLE, True):
            pass
        else: # show popup
            self.toggleTopicRead(event);
            # make popup menu
            mn = wx.wxMenu()
            self.addMenuItem(mn, 'Mark as &read', 'Mark this topic as read', self.markTopicRead)
            self.addMenuItem(mn, 'Mark as &unread', 'Mark this topic as unread', self.markTopicUnread)
            self.addMenuItem(mn, '&Delete', 'Remove this topic from the cache', self.deleteTopic)
            mn.AppendSeparator()
            self.addMenuItem(mn, '&Go to link', 'Go to the page linked in this topic', self.gotoLink)
            self.addMenuItem(mn, '&Copy', 'Copy title, summary and link on clipboard', self.copyTitleDescriptionLink)
            # show menu
            self.FeedTopics.PopupMenu(mn, event.GetPoint())
            # clean up
            for mi in mn.GetMenuItems():
                mn.Remove(mi.GetId())
                mi.Destroy()
            mn.Destroy()
        
    def toggleTopicRead(self, event):
        """Toggles the read/unread status of a topic."""
        # make sure it doesn't untoggle itself due to timer
        if self.ReadTimer.IsRunning:
            self.ReadTimer.Stop()
        try: # if right-clicked, wxGTK will not select item on its own, so do it programmatically
            self.FeedTopics.CurrentItem = event.m_itemIndex
            self.FeedTopics.SetItemState(self.FeedTopics.CurrentItem, wx.wxLIST_STATE_SELECTED, wx.wxLIST_STATE_SELECTED)            
            # ensure it doesn't auto-unselect again (there's still a bug though
            # if time to read is set to 0, right-clicking on a previously unselected
            # item will select it, but not toggle its read/unread)
            if self.ReadTimer.IsRunning:
                self.ReadTimer.Stop()
        except:
            pass
        if not self.validIds():
            return None
        status = engine.toggleItemReadUnread(self.FeedsList.CurrentItem, self.FeedTopics.CurrentItem)
        self.drawItem(self.FeedTopics, self.FeedTopics.CurrentItem, 
                      setToUnread=status,
                      setToCacheExceeded=self.FeedTopics.CurrentItem>=engine.getFeedCacheSize(self.FeedsList.CurrentItem))
        self.setReadStatusOfFeed(self.FeedsList.CurrentItem)
        self.resizeFeedsList()
        self.resizeFeedTopics()
        
    def validIds(self, checkTopicsId=True):
        """Returns True if both self.FeedTopics.CurrentItem
           and self.FeedsList.CurrentItem are integers and
           the selections are valid (i.e. not higher than
           the number of available items, False otherwise.
           If checkTopicsId is false, the validity of 
           self.FeedTopics.CurrentItem is not checked."""
        # first check types so it can fail immediately
        CorrectFeedsId = type(self.FeedsList.CurrentItem)==type(1) and \
                         self.FeedsList.CurrentItem>=0 and \
                         self.FeedsList.CurrentItem<self.FeedsList.GetItemCount()
        if checkTopicsId:
            CorrectTopicsId = type(self.FeedTopics.CurrentItem)==type(1) and \
                              self.FeedTopics.GetSelectedItemCount()>0 and \
                              self.FeedTopics.CurrentItem<self.FeedTopics.GetItemCount() and \
                              self.FeedTopics.CurrentItem>=0
        else:
            CorrectTopicsId = True
        return CorrectFeedsId and CorrectTopicsId
        
    def deleteTopic(self, event):
        """Deletes selected topic"""
        if not self.validIds():
            return None
        self.disableAll()
        if engine.deleteItem(self.FeedsList.CurrentItem, self.FeedTopics.CurrentItem):
            self.FeedTopics.DeleteItem(self.FeedTopics.CurrentItem)
            self.setReadStatusOfFeed(self.FeedsList.CurrentItem)
            savednumber = self.FeedTopics.CurrentItem
            self.displayItems(self.FeedsList.CurrentItem)
            self.FeedTopics.CurrentItem = savednumber
            try:
                self.FeedTopics.SetItemState(self.FeedTopics.CurrentItem, wx.wxLIST_STATE_SELECTED, wx.wxLIST_STATE_SELECTED)
            except:
                self.FeedTopics.CurrentItem = None
        self.enableAll()
        
    def gotoLink(self, event):
        """Go to the current link."""
        if not self.validIds():
            return None
        t = engine.getItem(self.FeedsList.CurrentItem, self.FeedTopics.CurrentItem)
        global url
        link = engine.getLink(t) or t.get("id", "")
        if link: url.open(link)
        
    def deleteAllTopics(self, event):
        """Deletes selected topic"""
        if not self.validIds(checkTopicsId=False):
            return None
        self.disableAll()
        try:
            temptext = 'You are about to delete all topics in feed \"%s\".\n' + \
                       'This action can not be undone. Proceed?'
            temptext = temptext % self.FeedsList.GetItem(self.FeedsList.CurrentItem).GetText()
            # ask confirmation
            ok = self.askConfirmation(amessage=temptext, acaption="Delete all topics in this feed?", 
                                      astyle=wx.wxYES_NO|wx.wxICON_QUESTION|wx.wxNO_DEFAULT)
            if ok:
                engine.deleteAllItemsInFeed(self.FeedsList.CurrentItem)
                self.displayItems(self.FeedsList.CurrentItem)
        finally:
            self.setReadStatusOfFeed(self.FeedsList.CurrentItem)
            self.enableAll()
            
    def textToClip(self, text):
        """Puts text on the clipboard."""
        textdata = wx.wxTextDataObject()
        textdata.SetText(text)
        clip = wx.wxTheClipboard
        clip.Open()
        clip.SetData(textdata)
        clip.Close()
            
    def copyLink(self, event):
        """Copies link in current topic to clipboard."""
        if not self.validIds(checkTopicsId=True):
            return None
        topic = engine.getItem(self.FeedsList.CurrentItem, self.FeedTopics.CurrentItem)
        link = engine.getLink(topic)
        if link:
            self.textToClip(link)
        else:
            print "This topic does not have a link!"
            
    def copyTitleDescriptionLink(self, event):
        """Copies title, description and link in current topic to clipboard."""
        if not self.validIds(checkTopicsId=True):
            return None
        topic = engine.getItem(self.FeedsList.CurrentItem, self.FeedTopics.CurrentItem)
        text = []
        title = topic.get("title", "")
        summary = topic.get("summary", "")
        content = engine.getContent(topic)
        link = engine.getLink(topic)
        
        if title:
            text.append(title)
        if summary:
            text.append(summary)
        if content:
            text.append(content)
        if link:
            text.append(link)
        if text:
            text = [ engine.htmlToPlain(t, DISPCODE) for t in text ]
            text = "\n".join(text)
            self.textToClip(text)
        else:
            print "This topic is invalid!"
        
    def copyAll(self, event):
        """Copies entire topic to clipboard."""
        if not self.validIds(checkTopicsId=True):
            return None
        topic = engine.getItem(self.FeedsList.CurrentItem, self.FeedTopics.CurrentItem)
        self.textToClip(self.topicAsText(topic))
            
    def markAllFeedsRead(self, event):
        """Marks all topics in all feeds as read"""
        engine.setAllFeedsRead(setToUnread=False)
        self.storeStatus()
        for feednumber in range(self.FeedsList.GetItemCount()):
            self.setReadStatusOfFeed(feednumber)
        if self.validIds(checkTopicsId=False):
            self.displayItems(self.FeedsList.CurrentItem)
        self.restoreStatus()
        
    def emptyAllFeeds(self, event):
        """Remove all topics from all feeds"""
        temptext = 'You are about to delete all topics in all feeds.' + \
                   'This action can not be undone. Proceed?'
        # ask confirmation
        ok = self.askConfirmation(amessage=temptext, acaption="Delete all topics in all feeds?", 
                                  astyle=wx.wxYES_NO|wx.wxICON_QUESTION|wx.wxNO_DEFAULT)
        if ok:
            engine.deleteAllItemsInAllFeeds()
            self.initFeeds()


def speCreate(parent, info=None):
    """Integration of Pears into spe (http://spe.pycs.net)"""
    global Pears
    global engine
    global mw
    global url
        
    Pears = MyFrameWithEvents(parent, -1, "")
        
    # create engine
    engine = pearsengine.engine()
        
    # only allow one instance, to avoid trouble with maintaining
    # integrity of the saved data
    try: # Win9x has problems with getpass
        import getpass
        currentuser = getpass.getuser()
    except:
        currentuser = "DEFAULT"
    checkone = wx.wxSingleInstanceChecker(Pears.GetAppName()+currentuser, engine.savedir)
    if checkone.IsAnotherRunning():
        sys.exit("Pears already running for user %s.\nIt is not allowed to run two instances for the same user." % getpass.getuser())
        
    # get the engine going
    engine.loadAll()
        
    engine.setSetting(TIMESRUN, engine.getSetting(TIMESRUN, 0)+1)
    url = urlOpener() # MUST start before the gui
    print url
        
    wx.wxInitAllImageHandlers()
    mw = MainFrame(None, -1, "")
    # make sure the window isn't tiny
    mw.doSize(None)
    mw.loadSettings()
        
    mw.initFeeds()
        
    mw.Show(1)
    return mw


def main():
    """Main function"""
    global Pears
    global engine
    global mw
    global url
        
    Pears = wx.wxPySimpleApp()
        
    # create engine
    engine = pearsengine.engine()
        
    # only allow one instance, to avoid trouble with maintaining
    # integrity of the saved data
    try: # Win9x has problems with getpass
        import getpass
        currentuser = getpass.getuser()
    except:
        currentuser = "DEFAULT"
    checkone = wx.wxSingleInstanceChecker(Pears.GetAppName()+currentuser, engine.savedir)
    if checkone.IsAnotherRunning():
        sys.exit("Pears already running for user %s.\nIt is not allowed to run two instances for the same user." % getpass.getuser())
            
        
    # get the engine going
    engine.loadAll()
        
    engine.setSetting(TIMESRUN, engine.getSetting(TIMESRUN, 0)+1)
    url = urlOpener() # MUST start before the gui
        
    wx.wxInitAllImageHandlers()
    mw = MainFrame(None, -1, "")
    # make sure the window isn't tiny
    mw.doSize(None)
    mw.loadSettings()
        
    mw.initFeeds()
        
    Pears.SetTopWindow(mw)
    mw.Show(1)
    Pears.MainLoop()


if __name__=="__main__":
    main()
