#!/usr/bin/env python

"""pearsconstants 1.7 - constants used by other parts of the program

Copyright (C) 2003, 2004 Project 5

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

Contact info
web: http://come.to/project5
mail/msn: project5@wanadoo.nl
icq: 84243714
jabber: project5@rhymbox.com
"""

# The "build" sets DEBUGMODE to False
# set to True in order to get extra debugging messages
DEBUGMODE = False


from wxPython import wx
WXPYTHONVERSION = wx.__version__

if WXPYTHONVERSION.endswith('u'):
    DISPCODE = None        # no encoding for Unicode-capable wxPython?
else:
    DISPCODE = 'iso8859-1' # encoding to use if wxPython is not Unicode?
#STORCODE = 'utf-8'     # format to store items in.  Not used!
PLAINCODE = 'utf-8'    # format to convert for hashing topic, guid etc!
FEEDCODE = 'cp1252'    # format to assume for converting numeric entities.
                       # FEEDCODE *must* evaluate to True in a boolean context,
                       # otherwise some feeds might break
OPMLCODE = 'utf-8' # encoding in OPML

# The LOCKFILEDIR is used by wxPython to create the 
# lock file in under Unix. This file is used to
# determine if there's only one instance of the app
# running. Override it in pearsdebug.py if Pears
# complains that the lock file already exists. The
# LOCKFILEDIR in pearsdebug.py can have any string 
# value as long as it evaluates to True in boolean 
# contexts.
LOCKFILEDIR = False

# constant identifiers of settings (used as settings names)
WIN_POS = "WindowPosition"
WIN_SIZE = "WindowSize"
FEEDSLIST_WIDTH = "ListFeedsWidth"
TOPICSLIST_HEIGHT = "ListTopicsHeight"
MS_READ = "msTillRead"
PLAINDRAWMODE = "PlainDrawMode"
STORECOLUMNS = "Restore column width"
SMARTDISPLAY = "Smart content displaying"
SHOWDIFF = "Show differences when updated"
COLUMNS = "Sizes of columns"
FEEDS = "Sizes of feeds columns"
TOPICS = "Sizes of topics columns"

TIMESRUN = "timesrun"
TOTALUPTIME = "Total uptime"


MAXLENGTH = 35 # maximum length of links
MAXTITLELENGTH = 60 # maximum length of topic titles
NEVER = 9999999 # do not change, or weird things might happen
RADIOPORT = 5335 # localhost port to listen for HTTP requests to add feeds
PEARS = "Pears"
SIMULTANEOUSUPDATE = 5 # maximum number of feeds updating at once

MAXFEEDSINMEMORY = SIMULTANEOUSUPDATE + 2 # maximum number of feeds in memory at any time
                                          # used by hamsterengine.py

AUTOUPDRESOLUTION =  2 # once every how many minutes to check for updates
                       # do NOT set lower than 1 or higher than 29!

# store location from where activated
import sys, os
PEARSLOCATION = os.path.split(sys.argv[0])[0] or os.getcwd() # *MUST* be the directory where everything, including About data and the likes are located
PYTHONVERSION = sys.version.split()[0]
import time
STARTTIME = time.time()


ID = "unique id"
PARENT = "parent id"
CHILDREN = "child ids"
SPECIAL = "special properties"
TOTAL = "total number of items"
TYPE = "type of item"
TFOLDER = "folder"
TFEED = "feed"
TNEWSITEM = "newsitem"

PREVIOUS = "previous item version"

LOCATION = "location"
TITLE = "title"
DATA = "downloaded data"
CACHE = CACHESIZE = "cache"
ETAG = "etag"
MODIFIED = "modified"
NEWITEMS = "__newitems" # used by hamsterengine.py
TOTITEMS = "__totitems" # used by hamsterengine.py
UPDATEINTERVAL = "auto-update interval in minutes"
LASTUPDATE = "datetime of last update"
DATETIME = "datetime when downloaded"

CHANNEL = "channel" # not necessary in version 2

HASHES = "hashes"
UNREAD = "unread"
ITEMS = "items"

# constants used as keys in the settings
TOPICS_AS_PLAIN_TEXT = "Return topics as plain text"
TOPICTEXT_REMOVE_IMG = "Remove images from topic text"
TOPICTEXT_MAKE_LINKS = "Make links for plain text URLs"
MAXLENGTH_LINKTEXT = "Maximum allowable link text length"
PARANOIDSAVING = "Paranoid saving mode"
REVERSEORDER = "Reverse order of downloaded items when storing"

FONT = "font in listctrls"
FFACE = "font face"
FSIZE = "font size"
FFAMILY = "font family"

BROWSERCOMMAND = "Browser startup command"

### CONSTANTS USED IN PLUGINS SYSTEM
PLUGINEXTENSION = '.pear'
PLongName = "LongName"
PShortName = "ShortName"
PHint = "Hint"
PVersion = "Version"
PAuthor = "Author"
PSource = "Source"
PClass = "Class"
PPlugins = "Plugins"
PUnknown = "[unknown]"
PStartPlugin = "### PEARSPLUGIN"
PStartCollection = "### PEARSPLUGINCOLLECTION"
PEnd = "### END"
PLUGINWARN = "Warn before running plugin"

PEARSRIGHTCLICKSTYLE = "Right-clicking behavior Pears style"

YES = 1
NO = 0
BACKUPEXT = "pzp" # extension of zipped backup files
MAKEBACKUPS = True # determines whether backups are made at startup
MAKECOMPRESSEDBACKUPS = False # determines whether backups are compressed (may be slow on old systems)

try: # allow overriding of these constants
    from pearsdebug import *
except:
    pass
