#!/usr/bin/env python

"""pearsengine 1.7 - back end for pears.py (gui-independant)

Copyright (C) 2003, 2004 Project 5

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

Contact info
web: http://come.to/project5
mail/msn: project5@wanadoo.nl
icq: 84243714
jabber: project5@rhymbox.com
"""

import os, os.path
import feedparser as rss
import md5
import formatter, htmllib, cgi
import copy
import sgmllib, re
import time, sys
from timeoutsocket import Timeout

__version__ = "Pears default (in-memory)"

from pearsconstants import *

class quasifile(list):
    """Behaves like a file as far as 
       formatter.DumbWriter is concerned."""
    def __init__(self, code, errors = 'replace'):
        self._code = code
        self._errors = errors
        
    def write(self, text):
        text = text.strip()
        if text:
            # if DEBUGMODE: print 'quasifile write', `text`
            if type(text) == type("s"):
                if self._code:
                    # string to string
                    self.append(text)
                else:
                    # string to unicode.
                    self.append(text.decode(FEEDCODE, 'replace'))
            elif type(text) == type(u"s"):
                #self.append(text.encode("ascii", "xmlcharrefreplace"))
                if self._code:
                    # unicode to string
                    self.append(text.encode(self._code, self._errors))
                else:
                    # unicode to unicode
                    self.append(text)
            else:
                # Hmm, would it be better to convert to unicode and then
                # encode with xmlcharreplace?  But we do not know the input
                # encoding that text is in!
                if DEBUGMODE: print 'cannot happen', `type(text)`
        
    def __nonzero__(self):
        # always evaluate as True
        return 1
        
    def astext(self):
        #if DEBUGMODE: print `self`        
        return " ".join(self)
        
    def reset(self):
        self.__delslice__(0, len(self))


import sgmllib, re


class LinkMaker(sgmllib.SGMLParser):
    """A parser which converts any stand-alone URL (meaning it is not
       inside a link, nor an attribute to a tag) to a proper link and
       can regenerate the modified HTML code.
       It also replaces IMG-tags with links to the images."""
    def __init__(self, makelinks=True, linkimages=True):
        """if makelinks==True, then links are created for URLS
           if linkimages==True, then image tags are converted to
             links to those images"""
        sgmllib.SGMLParser.__init__(self)
        self.makelinks = makelinks
        self.linkimages = linkimages
        self.anchorlevel = 0 # pays attention to nested anchors
        self.elements = [] # stores the bits and pieces which will be joined
        # don't allow generation of links which are huge. If the MAXLENGTH
        # is exceeded by an URL, it will be chopped a bit in self.makeLink.
        self.MAXLENGTH = MAXLENGTH
        # &'s must be replaced with some other code, as the parser chokes on 
        # them if they're in links. This code must be made out of chars and/or
        # digits in order to pass safely through the regex which registers
        # URL's. This character sequence is highly unlikely to appear anywhere
        # else.
        self.ampreplacement = "a1C2b3D4f5E6g7H8x9Z0yqW"
        
    def feed(self, text):
        """Does pre-processing on text before feeding it to the SGMLParser's
           feed method."""
        # replace ampersands with the ampreplacement
        text = text.replace("&", self.ampreplacement)
        # call the real feed method
        sgmllib.SGMLParser.feed(self, text)
        
    def unknown_starttag(self, tag, attrs):
        """Handle all start tags."""
        if tag == 'a':
            self.anchorlevel += 1 # store nested anchor depth
        if tag == 'img' and self.linkimages:
            # convert img-tag to link
            tag = 'a'
            # make a dictionary of attributes in order to convert those as well
            attribs = {"src": "", "alt": ""}
            for attrib in attrs:
                attribs[attrib[0]] = attrib[1].strip()
            # generate alt attribute from link if otherwise not available
            if not attribs["alt"]:
                attribs["alt"] = self.limitLength(attribs["src"])
            # only convert to link if a src attribute is present
            if attribs["src"]:
                self.elements.append('<a href="%s">%s (image)</a>' % \
                                     (attribs["src"], attribs["alt"]))
        else: # for non-img tags normal handling
            # build a string containing the attribs
            attribsstring = " ".join([ '%s=\"%s\"' % (attrib) for attrib in attrs ])
            if attribsstring:
                elem = " ".join([tag, attribsstring])
            else:
                elem = tag
            self.elements.append("<%s>" % elem)

    def unknown_endtag(self, tag):
        """Handle all end tags."""
        if tag == 'a':
            self.anchorlevel -= 1
            # don't allow anchorlevel <0 (invalid HTML in fact)
            self.anchorlevel = max(0, self.anchorlevel) 
        # convert img-tag to link
        if tag == 'img':
            tag = 'a'
        self.elements.append("</%s>" % tag)
        
    def handle_entityref(self, ref):
        """Is called when a character reference (&...) is found.
           These must pass through unmodified."""
        self.elements.append("&%s" % ref)
        
    def limitLength(self, text):
        """Returns a string with a maximum length of self.MAXLENGTH."""
        if len(text)>self.MAXLENGTH:
            # don't allow the text to become too large
            text = "%s...%s" % (text[:self.MAXLENGTH//2-2], 
                                text[-(self.MAXLENGTH//2-2):])
        return text
        
    def makeLink(self, matchobj):
        """Function called whenever linkify matches. Takes the
           match object and returns a link."""
        url = matchobj.group() # this will be in the href
        text = self.limitLength(url) # this will be the data of the tag (the visible link)
        return '<a href="%s">%s</a>' % (url, text)
        
    def linkify(self, text):
        """Regex for finding URLs:
           URL's start with http(s)/ftp/news ((http)|(ftp)|(news))
           followed by ://
           then any number of non-whitespace characters including
           numbers, dots, forward slashes, commas, question marks,
           ampersands, equality signs, dashes, underscores and plusses,
           but ending in a non-dot and non=plus!
           
           Result:
           (?:http|https|ftp|news)://(?:[@a-zA-Z0-9,/%:\&+#\?=\-_~;]+\.*)+[a-zA-Z0-9,/%:\&#\?=\-_]
           
           Tests:
              Plain old link: http://www.mail.yahoo.com. 
              Containing numbers: ftp://bla.com/di~ng/co.rt,39,%93 or other 
              Go to news://bl_a.com/?ha-h+a&query=tb for more info.
              A real link: <a href="http://x.com">http://x.com</a>. 
              ftp://verylong.org/url/must/be/chopped/to/pieces/oritwontfit.html (long one)
              <IMG src="http://b.com/image.gif" /> (a plain image tag)
              <a href=http://fixedlink.com/orginialinvalid.html>fixed</a> (original invalid HTML)
              Link containing an anchor <b>"http://myhomepage.com/index.html#01"</b>.
        """
        expression = r"(?:http|https|ftp|news)://(?:[a-zA-Z0-9,@/%:\&+#\?=\-_~;]+\.*)+[a-zA-Z0-9,/%:\&#\?\=\-_]"
        linkparser = re.compile(expression, re.I)
        text = linkparser.sub(self.makeLink, text)
        return text
        
    def handle_data(self, data):
        """Handle data between tags. If not inside a link
           (anchorlevel==0), then make sure any URLs are
           converted to links."""
        if self.anchorlevel==0 and self.makelinks:
            data = self.linkify(data)
        self.elements.append(data)
        
    def getResult(self):
        """Returns the updated HTML. This just consists of
           joining the elements."""
        result = "".join(self.elements)
        # replace ampreplacement with ampersands 
        return result.replace(self.ampreplacement, "&")


class engine(object):
    """Back-end doing the work."""
    def __init__(self):
        self.feeds = []
        self.items = {}
        self.cache = {} # contains stuff which has been downloaded, but not archived
        self.settings = {}
        self.savedir = self.getDefaultDir()
        self.parserversion = rss.__version__
        
    def getDefaultDir(self):
        """Gets location of default dir and creates it
           if necessary. ($HOME/.pears/)"""
        try:
            import pearsdebug
            savedir = pearsdebug.savedir
        except:
            savedir = os.path.expanduser(os.path.join('~', '.pears'))
            if len(savedir)<=len("c:\\/.pears"):
                # problem that might occur on Win2k (no $HOME environment variable)
                temp = os.path.join(os.path.expandvars('$USERPROFILE'), ".pears")
                if temp > len("c:\\/.pears"):
                    savedir = temp
            # create dir if it doesn't exist
            if not os.path.exists(savedir):
                os.makedirs(savedir)
        return savedir
        
    def restoreZippedBackup(self, afilename, wasdeleted):
        """Tries to restore afilename from the backup (<afilename>.pzp), 
           which is a zipped file.
           If wasdeleted==True, the program will show an error message saying
           that the file was deleted. Otherwise it will say it was corrupted.
           """
        result = None
        if wasdeleted:
            msg = "was not found. Perhaps it was deleted."
        else:
            msg = "can't be read. Perhaps it was corrupted."
        # first check if the file exists; otherwise, try to restore from backup
        backupname = ".".join([afilename, BACKUPEXT])
        if os.path.exists(backupname): # if there is a backup file
            try:
                import zipfile
                backupfile = zipfile.ZipFile(backupname, "r")
                try:
                    afn = os.path.split(afilename)[1] # the actual file name
                    if afn in backupfile.namelist(): # is the wanted file in the backup?
                        from pearsapp import ask
                        dorestore = ask("Restore backup?", 
                                        "\n".join(["The file:",
                                                   "   %s" % afilename,
                                                   "%s A backup was found in:" % msg,
                                                   "   %s" % backupname,
                                                   "",
                                                   "Do you wish to try to restore this backup?"
                                                  ]))
                        if dorestore == YES:
                            afile = file(afilename, "wb")
                            try:
                                t = backupfile.read(afn)
                                afile.write(t)
                                result = True
                            finally:
                                afile.close()
                finally:
                    backupfile.close()
            except:
                sys.excepthook(sys.exc_info()[0], sys.exc_info()[1], sys.exc_info()[2])            
                self.warn("Could not restore backup from:\n   %s" % backupname, "Backup restoration failed")
                result = None
        return result
        
    def makeZippedBackup(self, afilename):
        """Writes a zipped backup file of afilename to <afilename>.pzp."""
        backupname = ".".join([afilename, BACKUPEXT])
        try:
            import zipfile
            backupfile = zipfile.ZipFile(backupname, "w")
            try:
                if MAKECOMPRESSEDBACKUPS:
                    compression = zipfile.ZIP_DEFLATED
                else:
                    compression = zipfile.ZIP_STORED
                backupfile.write(afilename, os.path.split(afilename)[1], compression) # save it w/o the directory name
            finally:
                backupfile.close()
            return True
        except:
            sys.excepthook(sys.exc_info()[0], sys.exc_info()[1], sys.exc_info()[2])            
            return None
        
    def loadSomething(self, afilename, isbackup=False):
        """Loads a file, evals and returns that. Also makes backup of that
           file and tries to load a backupped version if the real file 
           has been removed.
           Wrapped by other load...() methods.
           
           isbackup determines whether the file this function is working
           on is already a backup copy (in which case it shouldn't be
           put back into the backup, nor should the program offer to 
           read the backup file if it's corrupted, since this IS the backup file).
           
           Returns None if failed."""
        if not os.path.exists(afilename) and not isbackup and MAKEBACKUPS:
            self.restoreZippedBackup(afilename, wasdeleted=True)
            isbackup = True
        if not os.path.exists(afilename):
            print "File %s does not exist" % afilename
            return None
        try:
            thefile = file(afilename, "r")
            line = thefile.readline()
            thefile.close()
        except:
            print "Could not load from:\n    %s" % afilename
            return None
        try:
            result = eval(line.replace('\r\n', '\n'))
            # if everything went ok (otherwise we'd be in the except right now), make a zipped backup
            if not isbackup and MAKEBACKUPS: # don't backup if it already is a backup version
                self.makeZippedBackup(afilename)
            return result
        except: # the program should never end up in this part of the code!
            ext = "ebak"
            warning = "\n".join([
                                "Could not evaluate contents of",
                                "   %s" % afilename,
                                "Pears will attempt to make a backup copy of this",
                                "file for examination purposes to",
                                "   %s.%s" % (afilename, ext),
                                "\nIf you do not know what caused this problem",
                                "(e.g. you didn't try to manipulate the file manually)",
                                "please report this and add the contents of your Pears",
                                "data directory to the report."]
                               )
            self.warn(warning)
            self.backup(afilename, ext)
            if not isbackup:
                self.restoreZippedBackup(afilename, False) # corrupted, perhaps the one in backup is still OK
                isbackup = True
                return self.loadSomething(afilename, isbackup)
            else:
                return None
            
    def warn(self, warning="", caption="Warning"):
        """Displays a warning message if possible in a wxPython dialog,
           otherwise in the console. Caption is used in the message header."""
        try:
            import pearsapp
            pearsapp.alert(caption, warning)
        except:
            print "\n".join([caption, warning])
            
    def backup(self, filename, ext):
        """Tries to make a backup of filename under the name
           <filename>.<ext>. Shows some warning messages if not
           successful, either in wxPython dialogs or in the console."""
        import shutil
        newname = ".".join([filename, ext])
        try:
            shutil.copyfile(filename, newname)
        except:
            self.warn("Could not make backup copy of\n   %s\nto\n   %s" % (filename, newname))
        
    def loadFeeds(self, filename=""):
        """Loads the list of feeds. Always call before other load...() methods.
           Wraps loadSomething()"""
        if filename.strip(): # load from parameter, otherwise keep default from __init__
            self.feedsfile = filename.strip()
            self.savedir = os.path.split(self.feedsfile)[0]
        # set up location of all necessary files    
        self.feedsfile = os.path.join(self.savedir, 'feeds.pyrs')
        self.itemsfile = os.path.join(self.savedir, 'items.pyrs')
        self.settingsfile = os.path.join(self.savedir, 'settings.pyrs')
            
        # load contents of feeds file (a list of descriptions+addresses)
        self.feeds = self.loadSomething(self.feedsfile)

        if self.feeds:
            for i in range(0, len(self.feeds)):
                # switch "description" to "tagline" for better compatibility with v3 of the feedparser
                # Code inserted in 1.7 beta 11
                if self.feeds[i].has_key("description"):
                    t = self.feeds[i]["description"]
                    del self.feeds[i]["description"]
                    self.feeds[i]["tagline"] = t

        if self.feeds==None:
            self.feeds = []
            # make default feed
            p5location = "http://project5.freezope.org/project5.rss"
            self.addFeed(title="Project 5", 
                         location=p5location,
                         channel={'rights': '&#169; Project 5', 
                                  'link': 'http://project5.tk', 
                                  'tagline': 'This feed contains mainly update announcements for the available software.', 
                                  'language': 'en', 
                                  'title': 'Project 5'},
                         saveimmediately=False
                        )
            newitem = {TITLE: "Welcome to Pears",
                       "link": "http://project5.freezope.org/pears",
                       "summary": \
                            """Pears is a 3-pane RSS/RDF newfeed aggregator. Using it is really simple:
                            add your own <b>feeds</b> to the list (<code><b>Ctrl+N</b></code> 
                            or <b><i>New</i></b> from the <b><i>Feeds</i></b> menu). You can also 
                            drag-and-drop links from your browser to the Pears window in order to add them.
                            <br /><br />
                            Each feed contains <b>topics</b>. Pears automatically caches
                            a number of those for each feed, so you don't have to be
                            online to read them. Note that if the cache size of a certain
                            feed is exceeded, some topics will be deleted next time that Pears
                            is shut down. The items which are to be deleted will be marked
                            using a different color. Feeds containing topics which will be 
                            deleted also have a different color.
                            <br /><br />
                            You can update an individual feed by <b>right-clicking</b> on 
                            it in the Feeds listbox (on the left) or by selecting it
                            and choosing <b><i>Update</i></b> from the <i>Feeds</i>
                            menu. You can update all your feeds at once using the 
                            <b><code>F5</code></b> key. Note that while updating a certain
                            feed, the name of that feed has a different markup in the feeds
                            list. If you have a slow connection or the feed is very large,
                            it might seem like Pears is frozen. This is <u>not</u> the case,
                            the interface will be updated again once the feed has been downloaded.
                            <br />
                            <br />
                            Each topic in a feed can have read/unread status. This is toggled
                            automatically when that topic is highlighted for a certain period
                            of time (at which point the topic is marked as Read). You can also
                            toggle it back and forth by right-clicking on the topic or by using
                            the <b><i>Topics</i></b> menu.
                            <br />
                            <br />
                            Topics can be entirely or partially copied to clipboard. For the 
                            different options, see the <i>Topics</i> menu.
                            <br />
                            <br />
                            Have a look at the About screen (<code><b>F2</b></code>) to view
                            details about the environment Pears is running on (useful for
                            backups, uninstalling and sending bug reports). More help is
                            available from the HTML docs (accessible in the Help menu, or
                            by pressing <code><b>F1</b></code>).
                            <br />
                            <br />
                            Feel free to <a href="mailto:project5@operamail.com">drop me a line</a>
                            if you like Pears, have a feature request or want to report a bug.
                            If you use Pears on a platform other than WinXP/Mandrake 9.1, I'd 
                            appreciate a screenshot as well.
                            """}
            self.items[p5location][ITEMS] = {'0': newitem}
            self.items[p5location][HASHES] = ['0']
            self.items[p5location][UNREAD] = ['0']
            # do NOT save immediately because it's possible that there is a corruption problem, in which
            # case the possibly intact items file will be overwritten
            #self.saveAll() 
            
    def loadItems(self):
        """Loads the items. Do not call before loadFeeds
           because it will not know where to get its stuff from."""
        # handle crash recovery a bit gracefully:
        # If the feeds file was absent (or corrupted), a self.items will have been generated 
        # containing a single "Welcome" entry in the project5 feed. If this is the case, make 
        # a backup of the items file (if present) in case the absence of the the feeds file
        # was due to a malfunction.
        if len(self.items) == 1: # if only one feed in the items dict
            key = self.items.keys()[0] # is the address of the feed
            if (len(self.items[key][HASHES])==1  # if one item in this feed
              and len(self.items[key][UNREAD])==1): # and that item is unread
                # now it's pretty much certain that the feeds file was not found/read
                # make a backup copy of the items file if possible
                if os.path.exists(self.itemsfile): 
                    # if it doesn't exist, it's most likely a new installation, so no warnings
                    warning = "\n".join(["It seems your feeds file was either corrupted or",
                                         "deleted! Pears will try to make a backup of your",
                                         "items.pyrs file. You can try and restore the feeds",
                                         "files from this using the rebuildfeeds.py utility",
                                         "in the Pears directory."] )
                    self.warn(warning)
                    self.backup(self.itemsfile, "rbak")
                self.saveAll()
        self.items = self.loadSomething(self.itemsfile)
        if self.items==None: self.items = {}
        # in case there was only a feeds file present and no items file,
        # make a bunch of default items
        self.checkFeeds()
        
    def checkFeeds(self):
        """Make sure that every feed has at least a valid empty items entry"""
        for feed in self.feeds:
            if not self.items.has_key(feed[LOCATION]):
                print "Feed at %s does not have any associated items. They will be created now." % feed[LOCATION]
                temp, emptyfeed = self.makeEmptyFeed()
                self.items[feed[LOCATION]] = emptyfeed
                # remove etag, since we lost state, need a full reload!
                feed[ETAG] = None
                
    def dumpEtag(self):
        """Resets all etag and modified entries in the feeds list in order 
           to force an update."""
        for feed in self.feeds:
            feed[ETAG] = None
            feed[MODIFIED] = None
            
    def loadSettings(self):
        """Loads the items. Do not call before loadFeeds
           because it will not know where to get its stuff from."""
        self.settings = self.loadSomething(self.settingsfile)
        if self.settings==None: self.settings = {}
            
    def saveSomething(self, something, afile):
        """Saves something to afile. Wrapped by other save...() methods."""
        try:
            thefile = file(afile, "w")
            thefile.write(str(something))
            thefile.write('\n')
            thefile.close()
        except:
            print "Could not save to:\n%s    " % afile
            
    def saveFeeds(self):
        """Saves the list of feeds to the feeds file.
           Wraps saveSomething()."""
        self.saveSomething(self.feeds, self.feedsfile)
        
    def saveItems(self):
        """Saves the items to the items file.
           Wraps saveSomething()."""
        self.saveSomething(self.items, self.itemsfile)
        
    def saveSettings(self):
        """Saves the settings to the settings file.
           Wraps saveSomething()."""
        self.saveSomething(self.settings, self.settingsfile)
        
    def saveAll(self):
        """Wraps saveFeeds() and saveItems()"""
        self.saveFeeds()
        self.saveItems()
        self.saveSettings()
        
    def loadAll(self):
        """Wraps loadFeeds() and loadItems()"""
        self.loadFeeds()
        self.loadItems()
        self.loadSettings()
        
    def getFeedsNames(self):
        """Return a list of all names of the available feeds"""
        return [ feed[TITLE] for feed in self.feeds ]
        
    def getFeedsLocations(self):
        """Return a list of all locations of the available feeds"""
        return [ feed[LOCATION] for feed in self.feeds ]
        
    def getFeedsForExport(self):
        """Return a list of tuples. Each tuple contains {location, title, description}
           of a feed."""
        locs = self.getFeedsLocations()
        titles = self.getFeedsNames()
        descriptions = [ feed[CHANNEL].get("tagline", "")
                         for feed in self.feeds ]
        return zip(locs, titles, descriptions)
        
    def convertid(self, id):
        """Converts an id (either an integer specifying an index in
           self.feeds or a string specifying an address) to a
           location (address) and returns that as well as the index:
               (location, index_in_feeds)
               
           If id is not in the feeds list, (None, None) is returned if
           an integer was passed or (id, None) if a string was passed.
           This peculiar behavior is very important, because otherwise
           things will break when sorting by unread item counts.
           """
        if type(id)==type(5): # is index
            try:
                return self.getFeedsLocations()[id], id
            except:
                print "Could not find item %s in feeds list (%s long)" % (id, len(self.getFeedsLocations()))
                return None, None
        elif type(id) in [type("s"), type(u"s")]: # is location
            if id in self.getFeedsLocations():
                index = self.getFeedsLocations().index(id)
                return id, index
            else: # do NOT change: will break sorting by unread items
                return id, None
        else:
            print "Invalid type for id (id = %s)" % id
            return None, None
            
    def downloadFeed(self, id, forcenew=True, etag=None, modified=None, 
                     agent=PEARS):
        """Download the feed specified by id (either an integer 
           specifying its index in self.feeds or a string
           specifying an address) and returns a dictionary
           of that feed.
           If possible, a cached copy is returned unless
           forcenew is True.
           """
        location, index = self.convertid(id)
        # only download if not cached or must be new
        if forcenew or (not self.cache.has_key(location)):
            # get etag and modified from list of feeds if available
            # and not force new
            if index>=0:
                if etag==None:
                    etag = self.feeds[index][ETAG]
                if modified==None:
                    modifed = self.feeds[index][MODIFIED]
            # cache the results
            try:
                temp = rss.parse(location, etag, modified, agent)
            except Timeout:
                temp = {}
                print "Timeout: feed at (%s) might not be online." % location
                
            enc = temp.get("encoding", FEEDCODE)
            if DEBUGMODE: print "enc:", enc
            # if enc: no point in iffing, because enc is retrieves using the get() method with default FEEDCODE
              
            def toUnicode(obj, enc):
                """Converts obj to unicode, assuming it currently is encoded in enc.
                   Calls itself recursively if obj is not a one-level depth object like a strign.
                   Returns unicodified obj.
                   
                   Arguments:
                     obj - string or list/dictionary containing strings encoded using
                           enc. The string or all string items inside the list/tuple/dict are
                           converted to unicode.
                     enc - encoding all string objects inside obj are in
                """
                if type(obj) == type(u""):
                    return obj
                try: # works if obj is a string or unicode object
                    obj = obj.decode(enc, "replace")
                    return obj
                except: # not a string; perhaps a dictionary/list?
                    #sys.excepthook(sys.exc_info()[0], sys.exc_info()[1], sys.exc_info()[2])
                    try: # try the dictionary approach
                        keys = obj.keys() # keys are always ASCII-encoded, so no point in converting them
                        for key in keys: # unicodify each value
                            obj[key] = toUnicode(obj[key], enc)
                        return obj
                    except: # not a dictionary or string; try list or tuple
                        #sys.excepthook(sys.exc_info()[0], sys.exc_info()[1], sys.exc_info()[2])
                        try: # tuple can't be modified in-place, so we'll simply turn sequence types into lists
                            if type(obj) in [type((1,2)), type([1,2])]:
                                c = []
                                for i in range(len(obj)):
                                    c.append(toUnicode(obj[i], enc))
                                return c
                        except: # must be something unsupported, like an integer
                            #sys.excepthook(sys.exc_info()[0], sys.exc_info()[1], sys.exc_info()[2])
                            if type(obj) not in [type(None), type(3), type(1.2)]:
                                print "ERROR (please report and include the address of the feed!): could not decode", str(`obj`)[:50]
                                if DEBUGMODE: print `obj.decode(enc, "replace")`
                            return obj
            temp["channel"] = toUnicode(temp["channel"], enc)
            temp["items"] = toUnicode(temp["items"], enc)
            
            # The "items" and "channel" keys contain duplicate content under different names, 
            # due to the way the feedparser works. These duplicates will be removed according 
            # to the table below. 
            doublemapper = [ "title_detail", # duplicated in title
                             "summary_detail", # duplicated in summary
                             "description", # duplicate of summary (for items) or tagline (for channel properties)
                             "content_encoded", # duplicated in content
                             "link", # duplicated in links
                             "category", # duplicated in categories
                             "guid", # duplicate of id
                           ]
            for key in doublemapper:
                if temp["channel"].has_key(key):
                    del temp["channel"][key]
            for item in temp["items"]:
                for key in doublemapper:
                    if item.has_key(key):
                        del item[key]
                # Don't allow summary and content to be (partially) the same.
                summary = item.get("summary", "")
                content = self.getContent(item)
                if content and summary:
                    # Check if there is overlap between summary and content. If there is, dump content and
                    # set summary to be the longest of the two texts
                    s = u"".join(summary.split())[5:min(25, len(summary)-10)].lower()
                    c = u"".join(content.split()).lower()
                    if c.find(s)>=0: # if the summary and the content overlap
                        if len(summary)<len(content):
                            item["summary"] = content
                        del item["content"]
                
            self.cache[location] = temp
                
            sc = self.cache[location]
            updatedone = False # stores whether the update has been successful
            # find out if the update has happened
            if sc.has_key("headers"):
                # in this case, it's an "online" feed
                if sc["headers"].has_key("date") or sc["headers"].has_key("server"):
                    # if connection was made, otherwise these keys are absent
                    updatedone = True
            else:
                # if not "online" feed, both 'items' and 'channel' will
                # be empty if the feed was not updated; if one of them
                # is present, the feed has been updated
                if sc["items"] or sc["channel"]:
                    updatedone = True
            if updatedone: 
                try:
                    self.feeds[index][LASTUPDATE] = time.time()
                except: # this is a new feed which has not yet been added to the list
                    if sc["channel"]:
                        sc["channel"][LASTUPDATE] = time.time()
                if DEBUGMODE:
                    print location, "updated"
            
            # while we're downloading anyway, let's update
            # the properties of the feed if it's in the self.feeds list
            try:
                if location in self.getFeedsLocations():
                    feedpos = self.getFeedsLocations().index(location)
                    if len(str(sc["channel"])) > 10: #only update if there's something there
                        self.feeds[feedpos][CHANNEL] = sc["channel"]
                    if sc.has_key("etag"):
                        self.feeds[feedpos][ETAG] = sc["etag"]
                    if sc.has_key("modified"):
                        self.feeds[feedpos][MODIFIED] = sc["modified"]
            except:
                print "Feed at location %s invalid." % location
        else: # if not forced and in cache
            sc = self.cache[location]
        return sc
        
    def downloadChannelProperties(self, id):
        """Returns the properties of the channel specified by id.
           Wraps downloadFeed()."""
        return self.downloadFeed(id)["channel"]
        
    def autoupdateRequired(self, id):
        """Calculates whether the time elapsed since LASTUPDATE of the
           feed exceeds the feed's UPDATEINTERVAL. Returns True if this
           is the case, False otherwise."""
        location, index = self.convertid(id)
        elapsed = time.time() - self.feeds[index][LASTUPDATE]
        # convert from seconds to minutes
        elapsed = elapsed // 60
        #print location, elapsed
        return elapsed > self.feeds[index][UPDATEINTERVAL]
        
    def getLink(self, item, n=0):
        """Returns the nth link from the links dictionary returned
           by version >2.6 of feedparser. Returns empty string if not found.
           
           Arguments:
             item - dictionary describing the item"""
        links = item.get("links", []) # list of dictionaries
        if not links: 
            return ""
        else:
            try:
                linkdict = links[n] # this dictionary contains a key "href" which is the actual link
                return linkdict.get("href", "")                 
            except: # n is probably invalid
                if DEBUGMODE: print "ERROR: getLink could not get link %s of item %s" % (n, `str(item)`)
                return ""
      
    def getContent(self, item, n=0):
        """Like getLink, only works on the content dictionary."""
        contents = item.get("content", []) # list of dictionaries
        if not contents: 
            return ""
        else:
            try:
                contentdict = contents[n] # this dictionary contains a key "value" which is the actual text
                return contentdict.get("value", "")                 
            except: # n is probably invalid
                if DEBUGMODE: print "ERROR: getContent could not get content %s of item %s" % (n, `str(item)`)
                return ""
        
    def updateFeed(self, id, doforce=True):
        """Updates the items of the channel specified by id.
           Wraps downloadFeed(). Do not call with an id which
           is not in the list of stored feeds!
           If doforce is False, then updating is only performed
           if LASTUPDATE was performed more than UPDATEINTERVAL 
           minutes ago, in other words if self.autoupdateRequired()
           returns True."""
        location, x = self.convertid(id)
        if not doforce: # if auto-updating, find out if update is necessary
            if not self.autoupdateRequired(x):
                #print "not updating", location
                return None # no update
        sc = self.downloadFeed(location, forcenew=True)
        newitems = sc["items"]
            
        if newitems:
            locationcounter = 0
            # fix bug with several identical items in one rss file leading to
            # that item being marked unread every time it's downloaded
            hashesthisrun = [] # lists of the hashes found in this latest download
            for item in newitems:
                # Hash system works like this:
                # If the item has both id (or guid) and title, a hash of these two is
                # used. Otherwise, if the item has both link and title, these
                # two are used. If that is not the case neither, the full
                # contents are hashed.
                # First get all identifiers which might be used here.
                guid = item.get("id", "")
                title = item.get("title", "")
                link = self.getLink(item)
                summary = item.get("summary", "")
                content = self.getContent(item)
                # now determine what to
                if guid and title:
                    tohash = u"".join([guid, title])
                elif link and title:
                    tohash = u"".join([link, title])
                else: # screwed up feed: no title/link/id used
                    tohash = u"".join([title, guid, link, summary, content])
                    if not summary and not content:
                        print "INFO: feed at %s contains weird items (no title and either id or link or description or content)" % location
                # Calculate identifying hash
                # before hashing everything must be converted to plain text
                # and whitespaces must be removed (in this order). Otherwise
                # feeds which randomly replace some characters with their &#...; 
                # equivalents, will always appear to contain new items                
                tohash = unicode(self.htmlToPlain(tohash).split())
                #tohash = u"".join(tohash)
                    
                hash = md5.new(tohash).hexdigest()
                hashesthisrun.append(hash)
                isnew = not hash in self.items[location][HASHES] 
                if not isnew and not hash in hashesthisrun: # check if item exists, but contents have been changed
                    currentitem = self.items[location][ITEMS][hash]
                    identical = True
                    currentsummary = currentitem.get("summary", "")
                    currentcontent = self.getContent(currentitem)
                    identical = (identical and 
                                u"".join(summary.split()) == u"".join(currentsummary.split()) )
                    identical = (identical and 
                                u"".join(content.split()) == u"".join(currentcontent.split()) )
                    if not identical: # then this is an updated item
                        # add current version to a "PREVIOUS" key
                        if currentitem.has_key(PREVIOUS): # dump the other historical stuff
                            del currentitem[PREVIOUS]
                        item[PREVIOUS] = copy.deepcopy(currentitem)
                        # remove item from the hashes
                        self.items[location][HASHES].remove(hash)
                        isnew = True # so that the item is processed anyway
                if isnew: # if new item indeed, then add
                    if self.getSetting(REVERSEORDER, False):
                        # if REVERSEORDER, then last item in the RSS is regarded as newest
                        # and is therefore put at the top
                        self.items[location][HASHES].insert(0, hash)
                    else:
                        self.items[location][HASHES].insert(locationcounter, hash)
                        locationcounter += 1
                    if hash not in self.items[location][UNREAD]:
                        self.items[location][UNREAD].append(hash)
                    self.items[location][ITEMS][hash] = item
            self.updateFeedCounts(location, x, doFlush=True)
            if self.getSetting(PARANOIDSAVING, False):
                self.saveAll()
            
    def getTopics(self, id, newonly=False):
        """Returns the topics in the channel specified by id.
           The topics are tuples of the form (text, isNew)
           If newonly is True, only unread topics are returned."""
        location, t = self.convertid(id)
        #~ # When sorting on read/unread, location/t will return
        #~ if location==None and type(id) in [type(''), type(u'')] and \
           #~ self.items.has_key(id):
            #~ location = id
        hashes = self.items[location][HASHES]
        result = []
        for hash in hashes:
            item = self.items[location][ITEMS][hash]
            if (hash in self.items[location][UNREAD]) or (not newonly):
                if item.has_key("title"):
                    text = self.htmlToPlain(item["title"], None)
                elif item.has_key("summary"):
                    text = self.htmlToPlain(item["summary"], None)
                elif item.has_key("link"):
                    text = self.htmlToPlain(item["link"], None)
                else:
                    text = "hash = %s (this is a bug in Pears, please report)" % hash
                if len(text)>MAXTITLELENGTH:
                    # make sure we don't break words
                    text = text[:MAXTITLELENGTH+5]
                    text = text.split()
                    if len(text)>1:
                        text = text[:-1]
                    text = " ".join(text)
                    text = text + "..."
                # make sure no HTML is included in the topic title
                result.append((text, hash in self.items[location][UNREAD]))
        return result
        
    def getTopicCounts(self, id):
        """Returns a tuple of (newtopics, totaltopics) for the
           channel specified by id. 
           Wraps getTopics()"""
        return len(self.getTopics(id, newonly=True)), len(self.getTopics(id))
        
    def getItem(self, feedid, topicid):
        """Returns the topic with topicid (position!) of
           the channel specified by feedid (location or
           position)."""
        location, tempvar = self.convertid(feedid)
        hash = self.items[location][HASHES][topicid]
        topic = copy.deepcopy(self.items[location][ITEMS][hash])
        #~ def modify(topic):
            #~ for key in topic.keys():
                #~ if key in ['link', u'link', 'guid', u'guid']:
                    #~ break # don't turn link text into anchors
                #~ if key not in [PREVIOUS]:
                    #~ topic[key] = self.modifyText(topic[key])
                #~ else:
                    #~ modify(topic[key])
        #~ modify(topic)
        return topic
        
    def setItemRead(self, feedid, topicid, setToUnread=False):
        """Sets the topic with topicid (position!) of
           the channel specified by feedid (location or
           position) as read."""
        try:
            location, tempvar = self.convertid(feedid)
            hash = self.items[location][HASHES][topicid]
            if (hash in self.items[location][UNREAD]) and (not setToUnread):
                self.items[location][UNREAD].remove(hash)
            elif (hash not in self.items[location][UNREAD]) and setToUnread:
                self.items[location][UNREAD].append(hash)
            self.updateFeedCounts(location, tempvar)
            if self.getSetting(PARANOIDSAVING, False):
                self.saveItems()
            return True
        except:
            print "setItemRead() failed on feedid=%s, topicid=%s, setToUnread=%s" % (feedid, topicid, setToUnread)
            return False
        
    def setAllItemsInFeedRead(self, feedid, setToUnread=False, doSave=True):
        """Sets all items in the channel specified by feedid 
           (location or position) as read (or unread if setToUnread)."""
        location, tempvar = self.convertid(feedid)
        if setToUnread: # add all hashes to unread list
            self.items[location][UNREAD] = self.items[location][HASHES][:]
        else: # remove all unread flags
            self.items[location][UNREAD] = []
        self.updateFeedCounts(location, tempvar)
        if doSave:
            if self.getSetting(PARANOIDSAVING, False):
                self.saveItems()
        return True
        
    def setAllFeedsRead(self, setToUnread=False):
        """Sets all items in all channels as read (or unread
           if setToUnread). Wraps setAllItemsInFeedRead()."""
        result = True
        for feednumber in range(len(self.feeds)):
            result = result and self.setAllItemsInFeedRead(feednumber, setToUnread, doSave=False)
        self.saveItems()
        return result
        
    def toggleItemReadUnread(self, feedid, topicid):
        """Changes the item status from unread to read or
           the other way around, depending on what its 
           current status is. Returns True if the new status 
           of the item is unread, False if it is read."""
        location, tempvar = self.convertid(feedid)
        hash = self.items[location][HASHES][topicid]
        if hash in self.items[location][UNREAD]:
            self.items[location][UNREAD].remove(hash)
            self.updateFeedCounts(location, tempvar)
            return False
        else:
            self.items[location][UNREAD].append(hash)
            self.updateFeedCounts(location, tempvar)
            return True
            
    def deleteItem(self, feedid, topicid):
        """Deletes an item from the feed."""
        try:
            location, tempvar = self.convertid(feedid)
            hash = self.items[location][HASHES][topicid]
            # remove vrom read/unread
            if hash in self.items[location][UNREAD]:
                self.items[location][UNREAD].remove(hash)
            # remove from items
            del self.items[location][ITEMS][hash]
            # finally, remove hash
            self.items[location][HASHES].remove(hash)
            self.updateFeedCounts(location, tempvar)
            return True
        except:
            print "deleteItem(%s, %s) failed" % (feedid, topicid)
            return False
            
    def deleteAllItemsInFeed(self, feedid):
        """Deletes all items"""
        location, tempvar = self.convertid(feedid)
        # faster than looping: make a new, empty self.items entry for this location
        self.items[location] = self.makeEmptyFeed()[1]
        self.updateFeedCounts(location, tempvar)
        
    def deleteAllItemsInAllFeeds(self):
        """Deletes all items in all feeds"""
        for i in range(0, len(self.feeds)):
            self.deleteAllItemsInFeed(i)
        
    def makeEmptyFeed(self, location="", title="", cachesize=0, 
                      channel={}, updateinterval=1440, # 1 day
                      etag=None, modified=None):
        """Returns a tuple of an empty feed and an empty items member."""
        return ({LOCATION:str(location), TITLE:title, 
                 CACHESIZE:cachesize, CHANNEL:channel,
                 ETAG:etag, MODIFIED:modified,
                 UPDATEINTERVAL:updateinterval,
                 LASTUPDATE:0, NEWITEMS:0, TOTITEMS:0},
                {HASHES:[], UNREAD:[], ITEMS:{}})
        
    def addFeed(self, location, title, cachesize=50, 
                channel={}, updateinterval=1440,
                etag=None, modified=None, saveimmediately=True):
        """Adds a feed based on its location (URL) and using the title,
           number of topics to store and the properties of the feed.
           Returns True if successful, False otherwise.
           If the location is already in the locations list, an update
           is performed.
           saveimmediately is used to determine whether the feeds are saved
           immediatly after adding the new feed. Is True in most cases, but
           for disaster recovery at startup it will be set to False.
           """
        if location in self.getFeedsLocations(): #update
            location, index = self.convertid(location)
            self.feeds[index][LOCATION] = str(location)
            self.feeds[index][TITLE] = title
            self.feeds[index][CACHESIZE] = cachesize
            self.feeds[index][UPDATEINTERVAL] = updateinterval
            if etag:
                self.feeds[index][ETAG] = etag
            if modified:
                self.feeds[index][MODIFIED] = modified
        else: # add new feed
            newfeed, newitems = self.makeEmptyFeed(location, title, cachesize, channel, updateinterval, etag, modified)
            self.feeds.append(newfeed)
            self.items[location] = newitems
            if location in self.cache.keys():
                self.updateFeed(location, doforce=False) # get items from cache
        if saveimmediately: 
            self.saveAll()
        return True
        
    def getFeed(self, id):
        """Returns the feed entry of self.feeds of the feed specified by id."""
        location, index = self.convertid(id)
        return self.feeds[index]
        
    def removeFeed(self, id):
        """Deletes feed at id (id can be either a string containing the
           location or an integer indicating the position of the item
           to be deleted in the list of feeds), plus any
           associated items. Returns False if not successful, True if
           successful."""
        location, index = self.convertid(id)
        self.unlockFeed(index)
        if index<0:
            return False
        # delete items
        if self.items.has_key(location):
            del self.items[location]
        # delete entry from feeds list
        entrynumber = self.getFeedsLocations().index(location)
        del self.feeds[entrynumber]
        if self.getSetting(PARANOIDSAVING, False):
            self.saveAll()
        return True
        
    def purgeItems(self):
        """Removes items beyond the item limit of every feed."""
        for feedid in range(0,len(self.feeds)):
            if not self.isFeedCacheExceeded(feedid):
                continue
            feed = self.feeds[feedid]
            location, cachesize = (feed[LOCATION], feed[CACHESIZE])
            # determine which hashes to keep and which to remove
            remove = self.items[location][HASHES][cachesize:]
            for hash in remove:
                # remove from unread
                if hash in self.items[location][UNREAD]:
                    self.items[location][UNREAD].remove(hash)
                # remove from items
                del self.items[location][ITEMS][hash]
                # remove hash from listing
                self.items[location][HASHES].remove(hash)
                self.updateFeedCounts(location, feedid)
            
    def isFeedCacheExceeded(self, id):
        """Determines if cache of feed at id (id can be either a string 
           containing the location or an integer indicating the position 
           of the item to be deleted in the list of feeds), is exceeded
           by the current number of items in memory in that feed.
           Returns False if cache not exceeded, True if exceeded."""
        location, index = self.convertid(id)
        return self.feeds[index][CACHESIZE]<len(self.items[location][HASHES])
        
    def getFeedCacheSize(self, id):
        """Returns how large the cache of feed at id (id can be either a 
           string containing the location or an integer indicating the 
           position of the item to be deleted in the list of feeds) is."""
        location, index = self.convertid(id)
        return self.feeds[index][CACHESIZE]
        
    def getTotalTopicsCount(self):
        """Returns the total number of topics."""
        total = 0
        for feed in self.feeds:
            total += len(self.items[feed[LOCATION]][HASHES])
        return total
        
    def getFeedsSize(self):
        """Returns the number of bytes occupied by the saved feeds."""
        return len(str(self.feeds))
        
    def getTotalTopicsSize(self):
        """Returns the number of bytes occupied by the saved items."""
        return len(str(self.items))
        
    def close(self):
        """Prepares engine for shutdown (purges items
           and saves everything)."""
        self.purgeItems()
        self.saveAll()
        
    def setSetting(self, settingname, value):
        """Sets settingname to value in self.settings."""
        self.settings[settingname] = value
        
    def getSetting(self, settingname, default=None):
        """Returns settingname from self.settings. If not found
           and default!=None, the default value is return *and*
           saved to the settings."""
        if self.settings.has_key(settingname):
            return self.settings[settingname]
        else:
            if default!=None:
                self.setSetting(settingname, default)
                return default
            else:
                return None
                
    def exportOpml(self, filename, feeds=None):
        """Saves the feeds list as OPML file. Returns 
           True if successful, False otherwise.
           If feeds==None, then all feeds are exported!"""
        if feeds==None:
            feeds = self.getFeedsForExport()
        #try:
        myfile = file(filename, "w")
        fw = myfile.write
        fw("""<?xml version="1.0" encoding="%s"?>\n""" % OPMLCODE)
        fw("""<opml>\n""")
        fw("""  <body>\n""")
        template = """    <outline title="%s" description="%s" xmlUrl="%s" htmlUrl="" />\n"""
        
        def prepare(s):
            """Returns a string ready for adding to the XML file."""
            # s = cgi.escape(self.htmlToPlain(s.replace("\"", "'")))
            try:
                s = cgi.escape(self.htmlToPlain(s.replace("\"", "'"),
                               OPMLCODE, 'xmlcharrefreplace'))
            except:
                s = ""
            return s
            
        for feed in feeds:
            _title = prepare(feed[1])
            _description = prepare(feed[2])
            _xmlUrl = prepare(feed[0])
            fw(template % (_title, _description, _xmlUrl,))
        fw("""  </body>\n""")
        fw("""</opml>\n""")
        myfile.close()
        return True
        #~ except IOError, (text):
            #~ print "ERROR: %s" % text
            #~ return False
        #~ except:
            #~ print "ERROR in engine.exportOpml()"
            #~ return False
            
    def importOpml(self, filename, nodupes=True):
        """Opens the file specified by filename and parses it
           as OPML (XML). This method returns a list of tuples,
           each tuple containing on index 0 the location, 
           on index 1 the tile and on index 2 the description.
           If nodupes==True, feeds which are already listed in
           the engine are filtered away."""
        # import necessary stuff; we do this here, because this function
        # is rarely used and there's no point in importing everything
        # at startup
        from xml.dom import minidom
        from xml.parsers.expat import ExpatError
        result = []
        try: 
            # try to parse the file
            xmldoc = minidom.parse(filename)
            try: # for some reason necessary for non-unicode wxPy
                enc = xmldoc.encoding
            except:
                enc = OPMLCODE
            feeds = xmldoc.getElementsByTagName("outline")
            # getval checks if the key exists and returns its value if it does,
            # or a default value if the key doesn't exist
            # IMPORTANT NOTE: do NOT use "has_key()" when manipulating XML,
            # as it apparently doesn't work in certain circumstances
            # (either a Linux issue, or an issue in older Python versions; not sure)
            getval = lambda feed, key, default="": \
                         (key in feed.attributes.keys()  #if key exists
                          and [feed.attributes[key].value]  # use its value
                          or [default] # otherwise the default value
                         )[0].strip() # "ternary" list trick -> take first elem; then strip
            result = [ ( getval(feed, "xmlUrl"),
                         getval(feed, "title", getval(feed,"xmlUrl")), 
                         getval(feed, "description", getval(feed, "title"))
                        )
                       for feed in feeds if "xmlUrl" in feed.attributes.keys() ]
        # now for some error handling
        except ExpatError, (text):
            print "ERROR: Invalid XML file! (%s)" % text
        except KeyError, (text):
            print "ERROR: %s" % text
        except IOError, (text):
            print "ERROR: %s" % text        
            
        # if necessary, filter out the feeds which are already in the engine
        if nodupes:
            existinglocs = self.getFeedsLocations()
            result = [ f for f in result if f[0] not in existinglocs ]
        return result
        
    def htmlToPlain(self, htmltext, code = PLAINCODE, errors = 'replace'):
        """Converts htmltext to plain text."""
        # DO NOT USE self.[something] in here, because it interferes with uniqueness
        # detection when using threads!
        # html parsing: generate everything
        qf = quasifile(code, errors)
        dumb = formatter.DumbWriter(file=qf) # plain text
        fmt = formatter.AbstractFormatter(dumb)
        plainparser = htmllib.HTMLParser(fmt)
        # reset everything
        #self.plainparser.reset()
        #self.quasifile.reset()
        # convert to plain text
        plainparser.feed(htmltext)
        plainparser.close()
        return qf.astext()
        
    def modifyText(self, text):
        """Applies desired modifications to the text."""
        # create the parser
        myparser = LinkMaker()
        # define what to do
        myparser.makelinks = self.getSetting(TOPICTEXT_MAKE_LINKS, True) # make links for plain text URLs
        myparser.linkimages = self.getSetting(TOPICTEXT_REMOVE_IMG, True) # make links out of img-tags
        myparser.MAXLENGTH = self.getSetting(MAXLENGTH_LINKTEXT, 65) # maximum allowable link text length when creating
        myparser.feed(text)
        return myparser.getResult()
        
    def lockFeed(self, *args, **kwargs):
        """Override in children"""
        pass
        
    def unlockFeed(self, *args, **kwargs):
        """Override in children"""
        pass
        
    def updateFeedCounts(self, location, id, doFlush = False):
        """Override in children"""
        pass
        
    def moveFeed(self, oldlocation, newlocation):
        """Tries to move a feed from oldlocation to newlocation.
           Method performs sanity checks on the parameters. 
           Returns new location of the feed (==oldlocation if
           move operation was not performed due to invalid
           parameters)."""
        try:
            feed = self.feeds.pop(oldlocation)
        except: # invalid oldlocation parameter
            if DEBUGMODE: print "engine.moveFeed() failed for oldlocation =", oldlocation
            return oldlocation
        # allow cyclic movement
        if newlocation > len(self.feeds):
            newlocation = 0 
        elif newlocation < 0:
            newlocation = len(self.feeds) # will be last feed
        self.feeds.insert(newlocation, feed)
        return newlocation
    
    def sortFeeds(self, inverse=False):
        """Sorts the feeds alphabetically (and in a case-insensitive
           fashion), based on their names. If inverse==True, the sorting
           is performed in inverse-alphabetical order."""
        def compnormal(a, b):
            a = a["title"]
            b = b["title"]
            a = u"".join(a.split()).lower()
            b = u"".join(b.split()).lower()
            return cmp(a, b)
        def compinv(a, b):
            a = a["title"]
            b = b["title"]
            a = u"".join(a.split()).lower()
            b = u"".join(b.split()).lower()
            return -cmp(a, b)
        if not inverse: comp = compnormal
        else: comp = compinv
        self.feeds.sort(comp)
        
    def sortFeedsUnread(self):
        """Sorts the feeds based on their unread status."""
        def comp(a, b):
            a = self.getTopicCounts(a[LOCATION])
            b = self.getTopicCounts(b[LOCATION])
            return -cmp(a, b)
        self.feeds.sort(comp)
        
    def reverseFeeds(self):
        """Reverse the order the feeds are currently in."""
        self.feeds.reverse()
        
    def randomizeFeeds(self):
        """Randomize the order the feeds are in - just in case the user
           is feeling a bit funny in the ole' head."""
        import random
        random.shuffle(self.feeds)
        
